#
# Known issues: Recentering on resize and when switching between
# different image types.  Ring centre on image switch.


from __future__ import annotations

import imp
import math
import os

import wx

from boost_adaptbx.boost.python import c_sizeof
from rstbx.viewer import image as rv_image
from rstbx.viewer import settings as rv_settings
from wxtbx import bitmaps

from ..rstbx_frame import EVT_EXTERNAL_UPDATE
from ..rstbx_frame import XrayFrame as XFBaseClass
from . import pyslip, tile_generation
from .calibration_frame import SBSettingsFrame

pyslip._Tiles = tile_generation._Tiles

# Use minimum value for an int to indicate a masked pixel
int_bits = c_sizeof("int") * 8
MASK_VAL = -(2 ** (int_bits - 1))


class chooser_wrapper:
    def __init__(self, image_set, index):
        self.image_set = image_set
        self.path = os.path.basename(image_set.get_path(index))
        self.full_path = image_set.get_path(index)
        self.index = index
        self._image_data = None

    def __str__(self):
        return "%s [%d]" % (self.path, self.index + 1)

    def get_detector(self):
        return self.image_set.get_detector()

    def get_scan(self):
        return self.image_set.get_scan()

    def get_beam(self):
        return self.image_set.get_beam()

    def get_mask(self):
        return self.image_set.get_mask(self.index)

    def get_image_data(self, corrected=True):
        if self._image_data is None:
            if corrected:
                return self.image_set.get_corrected_data(self.index)
            else:
                return self.image_set.get_raw_data(self.index)
        return self._image_data

    def set_image_data(self, image_data):
        self._image_data = image_data

    def get_detectorbase(self):
        return self.image_set.get_detectorbase(self.index)

    def get_vendortype(self):
        return self.image_set.get_vendortype(self.index)

    def show_header(self):
        return self.image_set.get_detectorbase(self.index).show_header()


class XrayFrame(XFBaseClass):
    def set_pyslip(self, parent):
        self.pyslip = pyslip.PySlip(parent, tile_dir=None, min_level=0)

    def __init__(self, *args, **kwds):
        self.params = kwds.get("params", None)
        if "params" in kwds:
            del kwds["params"]  # otherwise wx complains

        ### Collect any plugins
        slip_viewer_dir = os.path.join(os.path.dirname(__file__))
        contents = os.listdir(slip_viewer_dir)
        plugin_names = [
            f.split(".py")[0] for f in contents if f.endswith("_frame_plugin.py")
        ]
        self.plugins = {}
        for name in plugin_names:
            self.plugins[name] = imp.load_source(
                name, os.path.join(slip_viewer_dir, name + ".py")
            )
        if len(plugin_names) > 0:
            print("Loaded plugins: " + ", ".join(plugin_names))

        wx.Frame.__init__(self, *args, **kwds)
        self.settings = rv_settings()

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)

        # initialization is done in stages as windows are created
        self.pyslip = None
        self.viewer = wx.Panel(self, wx.ID_ANY)
        self.viewer.SetMinSize((640, 640))
        self.viewer.SetBackgroundColour(wx.BLACK)
        self.viewer.ClearBackground()
        self.sizer.Add(self.viewer, 1, wx.EXPAND)

        self.statusbar = self.CreateStatusBar()
        self.settings_frame = None
        self._calibration_frame = None
        self._ring_frame = None
        self._uc_frame = None
        self._score_frame = None
        self._plugins_frame = {key: None for key in self.plugins}
        self.zoom_frame = None
        self.plot_frame = None

        self.metrology_matrices = None

        # Currently displayed image.  XXX Can this be zapped?
        self._img = None

        self._distl = None
        self.toolbar = self.CreateToolBar(style=wx.TB_TEXT)
        self.setup_toolbar()
        self.toolbar.Realize()
        self.mb = wx.MenuBar()
        self.setup_menus()
        self.SetMenuBar(self.mb)
        self.Fit()
        self.SetMinSize(self.GetSize())
        self.SetSize((720, 720))

        self.Bind(EVT_EXTERNAL_UPDATE, self.OnExternalUpdate)

        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUICalibration, id=self._id_calibration)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUINext, id=wx.ID_FORWARD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIPrevious, id=wx.ID_BACKWARD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIRing, id=self._id_ring)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIUC, id=self._id_uc)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUIScore, id=self._id_score)
        for p in self.plugins:
            self.Bind(
                wx.EVT_UPDATE_UI,
                self.OnUpdateUIPluginWrapper(p),
                id=self._id_plugins[p],
            )

    # consolidate initialization of PySlip object into a single function
    def init_pyslip(self):
        self.set_pyslip(self.viewer)
        self.init_pyslip_presizer()

    def Show(self):
        # Due to the asynchronous nature of X11 on Linux, just showing a frame
        # does not guarantee window creation. The frame calls Raise() so that it
        # will be shown. This addresses an error with PySlip requiring the
        # window to exist before instantiation.
        super().Show()
        self.Raise()

    def setup_toolbar(self):
        XFBaseClass.setup_toolbar(self)

        btn = self.toolbar.AddTool(
            toolId=wx.ID_SAVEAS,
            label="Save As...",
            bitmap=bitmaps.fetch_icon_bitmap("actions", "save_all", 32),
            shortHelp="Save As...",
            kind=wx.ITEM_NORMAL,
        )
        self.Bind(wx.EVT_MENU, self.OnSaveAs, btn)

    # using StaticBox creates a horizontal white bar in Linux
    def make_gui(self, parent):
        parent.sizer = wx.BoxSizer(wx.HORIZONTAL)
        parent.SetSizer(parent.sizer)
        parent.sizer.Add(self.pyslip, 1, wx.EXPAND)

    def init_pyslip_presizer(self):
        self.demo_select_dispatch = {}

        # self.tile_directory = None#"/Users/nksauter/rawdata/demo/insulin_1_001.img"

        # build the GUI
        self.make_gui(self.viewer)

        # finally, bind event to handler
        self.pyslip.Bind(pyslip.EVT_PYSLIP_POSITION, self.handle_position_event)

    def handle_position_event(self, event):
        """Handle a pySlip POSITION event."""

        posn_str = ""
        if event.position:
            (lon, lat) = event.position
            fast_picture, slow_picture = self.pyslip.tiles.lon_lat_to_picture_fast_slow(
                lon, lat
            )

            posn_str = "Picture:  fast={:.3f} / slow={:.3f} pixels.".format(
                fast_picture + 0.5,
                slow_picture + 0.5,
            )
            coords = self.pyslip.tiles.get_flex_pixel_coordinates(lon, lat)
            if len(coords) >= 2:
                if len(coords) == 3:
                    readout = int(round(coords[2]))
                else:
                    readout = -1
                # the dials convention is that the center of the pixel is 0.5,0.5, so the extent
                # of the 0,0 pixel is from 0.0,0.0 to 1.0,1.0 in pixel space.  However
                # get_flex_pixel_coordinates defines the center of the pixel as 0.0,0.0, so we
                # are a half pixel off
                coords[0] += 0.5
                coords[1] += 0.5
                coords_str = f"fast={coords[1]:.3f} / slow={coords[0]:.3f} pixels"
                if len(coords) == 2:
                    posn_str += " Readout: " + coords_str + "."
                elif readout >= 0:
                    posn_str += " Readout %d: %s." % (readout, coords_str)
                possible_intensity = None
                fi = self.pyslip.tiles.raw_image
                detector = fi.get_detector()
                ifs = (int(coords[1]), int(coords[0]))  # int fast slow
                isf = (int(coords[0]), int(coords[1]))  # int slow fast
                image_data = fi.get_image_data()
                if not isinstance(image_data, tuple):
                    image_data = (image_data,)
                if readout >= 0:
                    if detector[readout].is_coord_valid(ifs):
                        possible_intensity = image_data[readout][isf]

                if possible_intensity is not None:
                    if possible_intensity == 0:
                        format_str = " I=%6.4f"
                        posn_str += format_str % possible_intensity
                    elif possible_intensity == MASK_VAL:
                        posn_str += " I=mask"
                    else:
                        yaya = int(math.ceil(math.log10(abs(possible_intensity))))
                        format_str = " I=%%6.%df" % (max(0, 5 - yaya))
                        posn_str += format_str % possible_intensity

                if (
                    len(coords) > 2 and readout >= 0
                ):  # indicates it's a tiled image in a valid region
                    reso = self.pyslip.tiles.get_resolution(
                        coords[1], coords[0], readout
                    )
                else:
                    reso = self.pyslip.tiles.get_resolution(coords[1], coords[0])

                if reso is not None:
                    posn_str += f" Resolution: {reso:.3f}"

            self.statusbar.SetStatusText(posn_str)
        else:
            self.statusbar.SetStatusText(
                "Click and drag to pan, mouse wheel or double click to zoom"
            )
            # print "event with no position",event
        return

    def init_pyslip_postsizer(self):
        self.pyslip.ZoomToLevel(-2)  # tiles.zoom_level
        self.pyslip.GotoPosition(
            self.pyslip.tiles.get_initial_instrument_centering_within_picture_as_lon_lat()
        )

    def setup_menus(self):
        file_menu = wx.Menu()
        self.mb.Append(file_menu, "File")
        item = file_menu.Append(-1, "Open integration results...")
        self.Bind(wx.EVT_MENU, self.OnLoadIntegration, item)
        item = file_menu.Append(-1, "Open image...")
        self.Bind(wx.EVT_MENU, self.OnLoadFile, item)
        self._actions_menu = wx.Menu()
        self.mb.Append(self._actions_menu, "Actions")
        # item = self._actions_menu.Append(-1, "Change beam center...")
        # self.Bind(wx.EVT_MENU, self.OnChangeBeamCenter, item)
        # item = self._actions_menu.Append(-1, "Reset beam center to header value")
        # self.Bind(wx.EVT_MENU, lambda evt: self.viewer.ResetBeamCenter(), item)
        item = self._actions_menu.Append(-1, "Save As...")
        self.Bind(wx.EVT_MENU, self.OnSaveAs, item)

        # Known wxWidgets/wxPython issue
        # (http://trac.wxwidgets.org/ticket/12394): stock item ID is
        # expected for zero-length text.  Work around by making text
        # contain single space. XXX Placement
        self._id_calibration = wx.NewId()
        item = self._actions_menu.Append(self._id_calibration, " ")
        self.Bind(wx.EVT_MENU, self.OnCalibration, source=item)

        # XXX Placement
        self._id_ring = wx.NewId()
        item = self._actions_menu.Append(self._id_ring, " ")
        self.Bind(wx.EVT_MENU, self.OnRing, source=item)

        # XXX Placement
        self._id_uc = wx.NewId()
        item = self._actions_menu.Append(self._id_uc, " ")
        self.Bind(wx.EVT_MENU, self.OnUC, source=item)

        # XXX Placement
        self._id_score = wx.NewId()
        item = self._actions_menu.Append(self._id_score, " ")
        self.Bind(wx.EVT_MENU, self.OnScore, source=item)

        self._id_plugins = {}
        for p in self.plugins:
            self._id_plugins[p] = wx.NewId()
            item = self._actions_menu.Append(self._id_plugins[p], " ")
            self.Bind(wx.EVT_MENU, self.OnPluginWrapper(p), source=item)

    def has_four_quadrants(self):
        d = self.pyslip.tiles.raw_image.get_detector()
        return len(d) > 1 and len(d.hierarchy()) == 4

    def add_file_name_or_data(self, file_name_or_data):
        """The add_file_name_or_data() function appends @p
        file_name_or_data to the image chooser, unless it is already
        present.  For file-backed images, the base name is displayed in
        the chooser.  If necessary, the number of entries in the chooser
        is pruned.  The function returns the index of the recently added
        entry.  XXX This is probably the place for heuristics to determine
        if the viewer was given a pattern, or a plain list of files.  XXX
        Rename this function, because it only deals with the chooser?
        """

        key = self.get_key(file_name_or_data)
        for i in range(self.image_chooser.GetCount()):
            if key == str(self.image_chooser.GetClientData(i)):
                return i
        if self.image_chooser.GetCount() >= self.CHOOSER_SIZE:
            self.image_chooser.Delete(0)
        i = self.image_chooser.GetCount()
        if type(file_name_or_data) is dict:
            self.image_chooser.Insert(key, i, None)
        elif isinstance(file_name_or_data, chooser_wrapper):
            self.image_chooser.Insert(key, i, file_name_or_data)
        else:
            self.image_chooser.Insert(os.path.basename(key), i, key)
        return i

    def get_beam_center_px(self):
        """
        Get the beam center in pixel coordinates relative to the tile closest to it.
        @return panel_id, beam_center_fast, beam_center_slow. panel_id is the panel the
        returned coordinates are relative to.
        """
        image = self.image_chooser.GetClientData(
            self.image_chooser.GetSelection()
        ).image_set
        detector = image.get_detector()
        beam = image.get_beam()
        if abs(detector[0].get_distance()) == 0:
            return 0.0, 0.0

        # FIXME assumes all detector elements use the same millimeter-to-pixel convention
        try:
            # determine if the beam intersects one of the panels
            panel_id, (x_mm, y_mm) = detector.get_ray_intersection(beam.get_s0())
        except RuntimeError as e:
            if not ("DXTBX_ASSERT(" in str(e) and ") failure" in str(e)):
                # unknown exception from dxtbx
                raise e
            # find the panel whose center is closest to the beam.
            panel_id = 0
            lowest_res = 0
            for p_id, panel in enumerate(detector):
                w, h = panel.get_image_size()
                res = panel.get_resolution_at_pixel(beam.get_s0(), (w // 2, h // 2))
                if res > lowest_res:
                    panel_id = p_id
                    lowest_res = res
            x_mm, y_mm = detector[panel_id].get_beam_centre(beam.get_s0())

        beam_pixel_fast, beam_pixel_slow = detector[panel_id].millimeter_to_pixel(
            (x_mm, y_mm)
        )

        return panel_id, beam_pixel_fast, beam_pixel_slow

    def load_image(self, file_name_or_data, get_image_data=None, show_untrusted=False):
        """The load_image() function displays the image from @p
        file_name_or_data.  The chooser is updated appropriately.
        """

        # Due to a bug in wxPython 3.0.2 for Linux
        # http://trac.wxwidgets.org/ticket/16034
        # the creation of the PySlip object is deferred until it is needed and
        # after other windows are created
        if self.pyslip is None:
            self.init_pyslip()
        # The settings dialog is created after PySlip because it may require
        # values from PySlip
        if self.settings_frame is None:
            self.OnShowSettings(None)
        self.Layout()

        if isinstance(file_name_or_data, chooser_wrapper):
            img = rv_image(file_name_or_data)
        else:
            try:
                img = rv_image(file_name_or_data.get_detectorbase())
            except AttributeError:
                img = rv_image(os.path.abspath(file_name_or_data))

        try:
            title = file_name_or_data.full_path
        except AttributeError:
            title = str(file_name_or_data)
        self.SetTitle(title)

        # Update the selection in the chooser.
        i = self.add_file_name_or_data(file_name_or_data)
        self.image_chooser.SetSelection(i)

        self.pyslip.tiles.show_untrusted = show_untrusted
        self.pyslip.tiles.current_brightness = self.settings.brightness
        self.pyslip.tiles.current_color_scheme = self.settings.color_scheme

        self.pyslip.tiles.set_image(
            file_name_or_data=img,
            metrology_matrices=self.metrology_matrices,
            get_image_data=get_image_data,
            show_saturated=(
                self.settings.display == "image"
                and self.settings.image_type == "corrected"
            ),
        )

        # Initialise position zoom level for first image.  XXX Why do we
        # have to coll ZoomToLevel to refresh subsequent images?
        if self._img is None:
            self.init_pyslip_postsizer()
        else:
            self.pyslip.ZoomToLevel(self.pyslip.tiles.zoom_level)

        self._img = img  # XXX

        self.update_statusbar()  # XXX Not always working?
        # self.Layout()

        detector = self.get_detector()
        if abs(detector[0].get_distance()) > 0:

            def map_coords(x, y, p):
                y, x = self.pyslip.tiles.flex_image.tile_readout_to_picture(
                    p, y - 0.5, x - 0.5
                )
                return self.pyslip.tiles.picture_fast_slow_to_map_relative(x, y)

            panel_id, beam_pixel_fast, beam_pixel_slow = self.get_beam_center_px()
            self.beam_center_cross_data = [
                (
                    (
                        map_coords(beam_pixel_fast + 3.0, beam_pixel_slow, panel_id),
                        map_coords(beam_pixel_fast - 3.0, beam_pixel_slow, panel_id),
                    ),
                    {"width": 2, "color": "#0000FFA0", "closed": False},
                ),
                (
                    (
                        map_coords(beam_pixel_fast, beam_pixel_slow + 3.0, panel_id),
                        map_coords(beam_pixel_fast, beam_pixel_slow - 3.0, panel_id),
                    ),
                    {"width": 2, "color": "#0000FFA0", "closed": False},
                ),
            ]

        # Unconditionally delete extra layers--update_settings() will add
        # them back if appropriate.  This also creates the self.*_layer
        # variables.
        if hasattr(self, "beam_layer") and self.beam_layer is not None:
            self.pyslip.DeleteLayer(self.beam_layer, update=False)
        self.beam_layer = None

        if hasattr(self, "spotfinder_layer") and self.spotfinder_layer is not None:
            self.pyslip.DeleteLayer(self.spotfinder_layer)
        self.spotfinder_layer = None

        if hasattr(self, "tile_layer") and self.tile_layer is not None:
            self.pyslip.DeleteLayer(self.tile_layer)
        self.tile_layer = None

        if hasattr(self, "tile_text_layer") and self.tile_text_layer is not None:
            self.pyslip.DeleteLayer(self.tile_text_layer)
        self.tile_text_layer = None

        # if hasattr(self, 'plugins_layer') and hasattr(self.plugins_layer, "__iter__"):
        #   for key in self.plugins_layer:
        #     if self.plugins_layer[key] is not None:
        #       self.pyslip.DeleteLayer(self.plugins_layer[key])
        # self.plugins_layer = {key:None for key in self.plugins}

        self.update_settings()

        # Destroy the calibration frame if it present but unsupported for
        # this image.  XXX Need to do something about the ring tool too
        # when switching between different kinds of images.  XXX Centering
        # is broken when switching between different kinds of images.
        if self._calibration_frame and not self.has_four_quadrants():
            self.OnCalibration(None)

    def get_detector(self):
        return self.pyslip.tiles.raw_image.get_detector()

    def get_beam(self):
        return self.pyslip.tiles.raw_image.get_beam()

    def get_key(self, file_name_or_data):
        """This overridden get_key() function returns the key of @p file_name_or_data
        if it's an DetectorImageBase object.  Otherwise it returns the super class's
        key
        """
        from iotbx.detectors.detectorbase import DetectorImageBase

        if isinstance(file_name_or_data, DetectorImageBase):
            return file_name_or_data.filename
        elif isinstance(file_name_or_data, chooser_wrapper):
            return str(file_name_or_data)
        else:
            return super().get_key(file_name_or_data)

    def update_settings(self, layout=True):
        raise NotImplementedError()

    def OnCalibration(self, event):
        if not self._calibration_frame:
            self._calibration_frame = SBSettingsFrame(
                self, wx.ID_ANY, "Quadrant calibration", style=wx.CAPTION | wx.CLOSE_BOX
            )
            self._calibration_frame.Show()
            self._calibration_frame.Raise()
        else:
            self._calibration_frame.Destroy()

    def OnRing(self, event):
        from .ring_frame import RingSettingsFrame

        if not self._ring_frame:
            self._ring_frame = RingSettingsFrame(
                self, wx.ID_ANY, "Ring tool", style=wx.CAPTION | wx.CLOSE_BOX
            )
            self._ring_frame.Show()
            self._ring_frame.Raise()
        else:
            self._ring_frame.Destroy()

    def OnUC(self, event):
        from .uc_frame import UCSettingsFrame

        if not self._uc_frame:
            self._uc_frame = UCSettingsFrame(
                self, wx.ID_ANY, "Unit cell tool", style=wx.CAPTION | wx.CLOSE_BOX
            )
            self._uc_frame.Show()
            self._uc_frame.Raise()
        else:
            self._uc_frame.Destroy()

    def OnScore(self, event):
        from .score_frame import ScoreSettingsFrame

        if not self._score_frame:
            self._score_frame = ScoreSettingsFrame(
                self, wx.ID_ANY, "Score tool", style=wx.CAPTION | wx.CLOSE_BOX
            )
            self._score_frame.Show()
            self._score_frame.Raise()
        else:
            self._score_frame.Destroy()

    def OnPluginWrapper(self, p):
        def OnPlugin(event):
            if not self._plugins_frame[p]:
                helper = self.plugins[p].PluginHelper
                self._plugins_frame[p] = helper._plugin_settings_frame(
                    self,
                    wx.ID_ANY,
                    helper._plugin_title,
                    style=wx.CAPTION | wx.CLOSE_BOX,
                )
                self._plugins_frame[p].Show()
                self._plugins_frame[p].Raise()
            else:
                self._plugins_frame[p].Destroy()

        return OnPlugin

    def OnUpdateUICalibration(self, event):
        # If quadrant calibration is not supported for this image, disable
        # the corresponding menu item.  Toggle the menu item text
        # depending on the state of the tool.

        if self.has_four_quadrants():
            event.Enable(True)
            if self._calibration_frame:
                event.SetText("Hide quadrant calibration")
            else:
                event.SetText("Show quadrant calibration")
        else:
            event.Enable(False)
            event.SetText("Show quadrant calibration")

    def OnUpdateUINext(self, event):
        # Enable/disable the "Next" button based on the image's position
        # in the list.

        event.Enable(
            self.image_chooser.GetSelection() + 1 < self.image_chooser.GetCount()
        )

    def OnUpdateUIPrevious(self, event):
        # Enable/disable the "Previous" button based on the image's
        # position in the list.

        event.Enable(self.image_chooser.GetSelection() >= 1)

    def OnUpdateUIRing(self, event):
        # Toggle the menu item text depending on the state of the tool.

        if self._ring_frame:
            event.SetText("Hide ring tool")
        else:
            event.SetText("Show ring tool")

    def OnUpdateUIUC(self, event):
        # Toggle the menu item text depending on the state of the tool.

        if self._uc_frame:
            event.SetText("Hide unit cell tool")
        else:
            event.SetText("Show unit cell tool")

    def OnUpdateUIScore(self, event):
        # Toggle the menu item text depending on the state of the tool.

        if self._score_frame:
            event.SetText("Hide score tool")
        else:
            event.SetText("Show score tool")

    def OnUpdateUIPluginWrapper(self, p):
        def OnUpdateUIPlugin(event):
            # Toggle the menu item text depending on the state of the tool.

            helper = self.plugins[p].PluginHelper
            if self._plugins_frame[p]:
                event.SetText(helper._plugin_hide_text)
            else:
                event.SetText(helper._plugin_show_text)

        return OnUpdateUIPlugin

    def OnSaveAs(self, event):
        ### XXX TODO: Save overlays
        ### XXX TODO: Fix bug where multi-asic images are slightly cropped due to transformation error'

        import PIL.Image as Image

        dialog = wx.FileDialog(
            self,
            defaultDir="",
            message="Save PNG or PDF file",
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT,
            wildcard="PNG file (*.png)|*.png|PDF file (*.pdf)|*.pdf",
        )
        if dialog.ShowModal() != wx.ID_OK:
            return

        file_name = dialog.GetPath()
        if file_name == "":
            return

        self.update_statusbar("Writing " + file_name + "...")
        if dialog.GetFilterIndex() == 0:
            # XXX Copied from tile_generation.py; all its disclaimers
            # apply.
            raw_img = self.pyslip.tiles.raw_image
            detector = raw_img.get_detector()
            data = raw_img.get_image_data()
            if not isinstance(data, tuple):  # XXX should not need this test
                data = (data,)
            from .tile_generation import get_flex_image_multipanel

            flex_img = get_flex_image_multipanel(
                brightness=self.settings.brightness / 100,
                detector=detector,
                image_data=data,
                beam=raw_img.get_beam(),
            )

            if flex_img.supports_rotated_tiles_antialiasing_recommended:
                currentZoom = self.pyslip.level
                self.pyslip.tiles.UseLevel(0)  # 1:1 zoom level

                try:
                    x, y, width, height = self._img._raw.bounding_box_mm()
                    x1, y1 = self._img._raw.detector_coords_as_image_coords(x, y)
                    x2, y2 = self._img._raw.detector_coords_as_image_coords(
                        x + width, y + height
                    )
                except AttributeError:
                    x1 = min(
                        [
                            p.get_pixel_lab_coord(c)[0] / p.get_pixel_size()[0]
                            for p in detector
                            for c in [
                                (0, 0),
                                (0, p.get_image_size()[1]),
                                (p.get_image_size()[0], 0),
                                (p.get_image_size()[0], p.get_image_size()[1]),
                            ]
                        ]
                    )
                    y1 = min(
                        [
                            p.get_pixel_lab_coord(c)[1] / p.get_pixel_size()[1]
                            for p in detector
                            for c in [
                                (0, 0),
                                (0, p.get_image_size()[1]),
                                (p.get_image_size()[0], 0),
                                (p.get_image_size()[0], p.get_image_size()[1]),
                            ]
                        ]
                    )
                    x2 = max(
                        [
                            p.get_pixel_lab_coord(c)[0] / p.get_pixel_size()[0]
                            for p in detector
                            for c in [
                                (0, 0),
                                (0, p.get_image_size()[1]),
                                (p.get_image_size()[0], 0),
                                (p.get_image_size()[0], p.get_image_size()[1]),
                            ]
                        ]
                    )
                    y2 = max(
                        [
                            p.get_pixel_lab_coord(c)[1] / p.get_pixel_size()[1]
                            for p in detector
                            for c in [
                                (0, 0),
                                (0, p.get_image_size()[1]),
                                (p.get_image_size()[0], 0),
                                (p.get_image_size()[0], p.get_image_size()[1]),
                            ]
                        ]
                    )

                # Map > View - determine layout in X direction
                x_offset = x1
                start_x_tile = int(math.floor(x_offset / self.pyslip.tile_size_x))
                stop_x_tile = (
                    x2 + self.pyslip.tile_size_x - 1
                ) / self.pyslip.tile_size_x
                stop_x_tile = int(stop_x_tile)
                col_list = range(start_x_tile, stop_x_tile)
                x_pix = start_x_tile * self.pyslip.tile_size_y - x_offset

                y_offset = y1
                start_y_tile = int(math.floor(y_offset / self.pyslip.tile_size_y))
                stop_y_tile = (
                    y2 + self.pyslip.tile_size_y - 1
                ) / self.pyslip.tile_size_y
                stop_y_tile = int(stop_y_tile)
                row_list = range(start_y_tile, stop_y_tile)
                y_pix_start = start_y_tile * self.pyslip.tile_size_y - y_offset

                bitmap = wx.Bitmap(x2 - x1, y2 - y1)
                dc = wx.MemoryDC()
                dc.SelectObject(bitmap)

                # start pasting tiles
                for x in col_list:
                    y_pix = y_pix_start
                    for y in row_list:
                        dc.DrawBitmap(
                            self.pyslip.tiles.GetTile(x, y), x_pix, y_pix, False
                        )
                        y_pix += self.pyslip.tile_size_y
                    x_pix += self.pyslip.tile_size_x

                dc.SelectObject(wx.NullBitmap)

                wximg = bitmap.ConvertToImage()
                imageout = Image.new("RGB", (wximg.GetWidth(), wximg.GetHeight()))
                imageout.frombytes(bytes(wximg.GetData()))

                self.pyslip.tiles.UseLevel(currentZoom)

            else:  # write the image out at full resolution
                flex_img.setWindow(0.0, 0.0, 1)
                flex_img.spot_convention(0)
                flex_img.adjust(color_scheme=self.settings.color_scheme)
                flex_img.prep_string()
                data_string = flex_img.as_bytes()
                imageout = Image.frombytes(
                    "RGB", (flex_img.ex_size2(), flex_img.ex_size1()), data_string
                )

            with open(file_name, "wb") as fh:
                imageout.save(fh, "PNG")

        elif dialog.GetFilterIndex() == 1:
            from reportlab.lib.units import inch
            from reportlab.pdfgen import canvas

            # Dots per inch in PDF output, and fudge factor to not make
            # fine features impossibly small.  XXX The fudge factor should
            # go.
            DPI = 72
            LINE_WIDTH_FACTOR = 0.6

            # XXX Copied from tile_generation.py; all its disclaimers
            # apply.
            raw_img = self.pyslip.tiles.raw_image
            detector = raw_img.get_detector()
            data = raw_img.get_image_data()
            if not isinstance(data, tuple):  # XXX should not need this test
                data = (data,)
            from .tile_generation import get_flex_image_multipanel

            flex_img = get_flex_image_multipanel(
                brightness=self.settings.brightness / 100,
                detector=detector,
                image_data=data,
                beam=raw_img.get_beam(),
            )

            flex_img.setWindow(0, 0, 1)
            flex_img.adjust(color_scheme=self.settings.color_scheme)
            flex_img.prep_string()

            # XXX Order of size1/size2 correct?
            pdf_size = (flex_img.size2() * inch / DPI, flex_img.size1() * inch / DPI)
            pdf_canvas = canvas.Canvas(filename=file_name, pagesize=pdf_size)
            pil_img = Image.frombytes(
                "RGB", (flex_img.size2(), flex_img.size1()), flex_img.as_bytes()
            )

            pdf_canvas.drawInlineImage(
                pil_img, 0, 0, width=pdf_size[0], height=pdf_size[1]
            )

            for layer_id in self.pyslip.layer_z_order:
                layer = self.pyslip.layer_mapping[layer_id]

                # XXX This would probably be more elegant if these were
                # functions in some layer class.  Note repeated roundabout
                # way (via a wx.Pen object) to extract RGB values from the
                # colour parameter.
                if layer.type == self.pyslip.TypeEllipse:
                    for (
                        p,
                        place,
                        width,
                        colour,
                        closed,
                        filled,
                        fillcolour,
                        x_off,
                        y_off,
                        pdata,
                    ) in layer.data:
                        if layer.map_rel:
                            pp = []
                            for pelement in p:
                                fs = (
                                    self.pyslip.tiles.map_relative_to_picture_fast_slow(
                                        pelement[0], pelement[1]
                                    )
                                )
                                pp.append(
                                    (
                                        fs[0] * inch / DPI,
                                        pdf_size[1] - fs[1] * inch / DPI,
                                    )
                                )
                            ellipse_center = (pp[0][0], pp[0][1])
                            major = (
                                pp[1][0] - ellipse_center[0],
                                pp[1][1] - ellipse_center[1],
                            )
                            minor = (
                                pp[2][0] - ellipse_center[0],
                                pp[2][1] - ellipse_center[1],
                            )
                        else:
                            raise NotImplementedError(
                                "PDF output in view-relative coordinates not implemented"
                            )

                        pen = wx.Pen(colour)
                        pdf_canvas.setLineWidth(width * LINE_WIDTH_FACTOR)
                        pdf_canvas.setStrokeColorRGB(
                            pen.Colour.Red() / 255,
                            pen.Colour.Green() / 255,
                            pen.Colour.Blue() / 255,
                        )

                        angle = math.atan2(major.elems[1], major.elems[0])
                        r_major = math.hypot(major.elems[0], major.elems[1])
                        r_minor = math.hypot(minor.elems[0], minor.elems[1])

                        pdf_canvas.saveState()
                        pdf_canvas.translate(
                            ellipse_center.elems[0], ellipse_center.elems[1]
                        )

                        pdf_canvas.rotate(math.degrees(angle))
                        pdf_canvas.ellipse(-r_major, -r_minor, r_major, r_minor)
                        pdf_canvas.restoreState()

                elif layer.type == self.pyslip.TypeImage:
                    raise NotImplementedError(
                        "PDF output of image layers not implemented"
                    )

                elif layer.type == self.pyslip.TypePoint:
                    for (
                        lon,
                        lat,
                        place,
                        radius,
                        colour,
                        x_off,
                        y_off,
                        pdata,
                    ) in layer.data:
                        if layer.map_rel:
                            fs = self.pyslip.tiles.map_relative_to_picture_fast_slow(
                                lon, lat
                            )
                        else:
                            raise NotImplementedError(
                                "PDF output in view-relative coordinates not implemented"
                            )

                        pt = (fs[0] * inch / DPI, pdf_size[1] - fs[1] * inch / DPI)
                        pen = wx.Pen(colour)
                        pdf_canvas.setLineWidth(radius)
                        pdf_canvas.setStrokeColorRGB(
                            pen.Colour.Red() / 255,
                            pen.Colour.Green() / 255,
                            pen.Colour.Blue() / 255,
                        )
                        pdf_canvas.circle(pt[0], pt[1], 0.5 * radius * inch / DPI)

                elif layer.type == self.pyslip.TypePolygon:
                    for (
                        p,
                        place,
                        width,
                        colour,
                        closed,
                        filled,
                        fillcolour,
                        x_off,
                        y_off,
                        pdata,
                    ) in layer.data:
                        path = pdf_canvas.beginPath()
                        for i, pp in enumerate(p):
                            if layer.map_rel:
                                fs = (
                                    self.pyslip.tiles.map_relative_to_picture_fast_slow(
                                        pp[0], pp[1]
                                    )
                                )
                            else:
                                raise NotImplementedError(
                                    "PDF output in view-relative coordinates not implemented"
                                )

                            pt = (fs[0] * inch / DPI, pdf_size[1] - fs[1] * inch / DPI)
                            if i == 0:
                                path.moveTo(pt[0], pt[1])
                            else:
                                path.lineTo(pt[0], pt[1])

                        if closed:
                            path.close()

                        pen = wx.Pen(colour)
                        pdf_canvas.setFillColorRGB(
                            pen.Colour.Red() / 255,
                            pen.Colour.Green() / 255,
                            pen.Colour.Blue() / 255,
                        )
                        pdf_canvas.setLineWidth(width * LINE_WIDTH_FACTOR)
                        pdf_canvas.setStrokeColorRGB(
                            pen.Colour.Red() / 255,
                            pen.Colour.Green() / 255,
                            pen.Colour.Blue() / 255,
                        )
                        pdf_canvas.drawPath(path, fill=filled)

                elif layer.type == self.pyslip.TypeText:
                    for (
                        lon,
                        lat,
                        tdata,
                        placement,
                        radius,
                        colour,
                        textcolour,
                        fontname,
                        fontsize,
                        offset_x,
                        offset_y,
                        data,
                    ) in layer.data:
                        if placement != "cc":
                            print(
                                Warning(
                                    "Only centered placement available when drawing text on pdf"
                                )
                            )
                        if layer.map_rel:
                            fs = self.pyslip.tiles.map_relative_to_picture_fast_slow(
                                lon, lat
                            )
                        else:
                            raise NotImplementedError(
                                "PDF output in view-relative coordinates not implemented"
                            )
                        from reportlab.pdfbase.pdfmetrics import stringWidth

                        scale = 5  # XXX this scaleup by 5 is arbitrary!
                        try:
                            w = stringWidth(tdata, fontname, fontsize * scale)
                        except KeyError:
                            fontname = "Helvetica"
                            w = stringWidth(tdata, fontname, fontsize * scale)
                        if fs[0] - (w / 2) < 0:  # handle text falling off the left side
                            txt = pdf_canvas.beginText(x=0, y=fs[1])
                        else:
                            txt = pdf_canvas.beginText(x=fs[0] - (w / 2), y=fs[1])
                        txt.setFont(fontname, fontsize * scale)
                        if isinstance(textcolour, wx.Colour):
                            textcolour = tuple(textcolour)
                        txt.setFillColor(textcolour)
                        txt.setStrokeColor(textcolour)
                        txt.textLine(tdata)
                        pdf_canvas.drawText(txt)

            pdf_canvas.save()

        self.update_statusbar("Writing " + file_name + "..." + " Done.")



"""Contains classes for refinement engines. Refinery is the shared interface,
LevenbergMarquardtIterations, GaussNewtonIterations, SimpleLBFGS and LBFGScurvs
are the current concrete implementations"""

from __future__ import annotations

import copy
import json
import logging
from io import StringIO
from typing import List, Union

import libtbx
from libtbx import easy_mp
from libtbx.phil import parse
from scitbx import lbfgs, sparse
from scitbx.array_family import flex
from scitbx.lstbx import normal_eqns, normal_eqns_solving

from dials.algorithms.refinement import DialsRefineRuntimeError

from .target import Target

logger = logging.getLogger(__name__)


# termination reason strings
TARGET_ACHIEVED = "RMSD target achieved"
RMSD_CONVERGED = "RMSD no longer decreasing"
STEP_TOO_SMALL = "Step too small"
OBJECTIVE_INCREASE = "Refinement failure: objective increased"
MAX_ITERATIONS = "Reached maximum number of iterations"
MAX_TRIAL_ITERATIONS = "Reached maximum number of consecutive unsuccessful trial steps"
DOF_TOO_LOW = "Not enough degrees of freedom to refine"
OBJECTIVE_FLAT = "Refinement failure: zero expected decrease of the objective"


refinery_phil_str = """
refinery
  .help = "Parameters to configure the refinery"
  .expert_level = 1
{
  engine = SimpleLBFGS LBFGScurvs GaussNewton *LevMar SparseLevMar
    .help = "The minimisation engine to use"
    .type = choice

  max_iterations = None
    .help = "Maximum number of iterations in refinement before termination."
            "None implies the engine supplies its own default."
    .type = int(value_min=1)

  log = None
    .help = "Filename for an optional log that a minimisation engine may use"
            "to write additional information"
    .type = path

  journal
    .help = "Extra items to track in the refinement history"
  {
    track_step = False
      .help = "Record parameter shifts history in the refinement journal, if"
              "the engine supports it."
      .type = bool

    track_gradient = False
      .help = "Record parameter gradients history in the refinement journal, if"
              "the engine supports it."
      .type = bool

    track_parameter_correlation = False
      .help = "Record correlation matrix between columns of the Jacobian for"
              "each step of refinement."
      .type = bool

    track_jacobian_structure = False
      .help = "Record numbers of explicit and structural zeroes in each column"
              "of the Jacobian at each step of refinement."
      .type = bool

    track_condition_number = False
      .help = "Record condition number of the Jacobian for each step of "
              "refinement."
      .type = bool

    track_normal_matrix = False
      .help = "Record the full normal matrix at each step of refinement"
      .type = bool

    track_out_of_sample_rmsd = False
      .type = bool
      .help = "Record RMSDs calculated using the refined experiments with"
              "reflections not used in refinement at each step. Only valid if a"
              "subset of input reflections was taken for refinement"
  }
}
"""
refinery_phil_scope = parse(refinery_phil_str)


class Journal(dict):
    """Container in which to store information about refinement history.

    This is simply a dict but provides some extra methods for access that
    maintain values as columns in a table. Refinery classes will use these methods
    while entering data to ensure the table remains consistent. Methods inherited
    from dict are not hidden for ease of use of this object when returned to the
    user."""

    reason_for_termination = None
    _nrows = 0

    def get_nrows(self):
        return self._nrows

    def add_column(self, key):
        """Add a new column named by key"""
        self[key] = [None] * self._nrows

    def add_row(self):
        """Add an element to the end of each of the columns. Fail if any columns
        are the wrong length"""

        for k in self:
            assert len(self[k]) == self._nrows
            self[k].append(None)
        self._nrows += 1

    def del_last_row(self):
        """Delete the last element from the each of the columns. Fail if any columns
        are the wrong length"""

        if self._nrows == 0:
            return None
        for k in self:
            assert len(self[k]) == self._nrows
            self[k].pop()
        self._nrows -= 1

    def set_last_cell(self, key, value):
        """Set last cell in column given by key to value. Fail if the column is the
        wrong length"""

        assert len(self[key]) == self._nrows
        self[key][-1] = value

    def to_json_file(self, filename):
        d = {"attributes": self.__dict__, "data": dict(self)}
        with open(filename, "w") as f:
            json.dump(d, f)

    @classmethod
    def from_json_file(cls, filename):
        with open(filename) as f:
            d = json.load(f)
        j = cls()
        j.update(d["data"])
        for key in d["attributes"]:
            setattr(j, key, d["attributes"][key])
        return j


class Refinery:
    """Interface for Refinery objects. This should be subclassed and the run
    method implemented."""

    # NOTES. A Refinery is initialised with a Target function. The target
    # function already contains a ReflectionManager (which holds the data) so
    # there's no need to pass the data in here. In fact the Target
    # class does the bulk of the work, as it also does the reflection prediction
    # to get the updated predictions on each cycle. This should make some sense
    # as the target function is inextricably linked to the space in which
    # predictions are made (e.g. detector space, phi), so it is not general
    # enough to sit abstractly above the prediction.

    # This keeps the Refinery simple and able to be focused only on generic
    # features of managing a refinement run, like reporting results and checking
    # termination criteria.

    # The prediction values come from a PredictionParameterisation object.
    # This is also referred to by the Target function, but it makes sense for
    # Refinery to be able to refer to it directly. So refinery should keep a
    # separate link to its PredictionParameterisation.

    def __init__(
        self,
        target: Target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=None,
    ):

        # reference to PredictionParameterisation, Target and ConstraintsManager
        # objects
        self._parameters = prediction_parameterisation
        self._target = target
        self._constr_manager = constraints_manager

        # initial parameter values
        self.x = flex.double(self._parameters.get_param_vals())
        if self._constr_manager is not None:
            self.x = self._constr_manager.constrain_parameters(self.x)
        self.old_x = None

        # undefined initial functional and gradients values
        self._f = None
        self._g = None
        self._jacobian: Union[flex.double, sparse.matrix, None] = None

        # filename for an optional log file
        self._log = log

        self._target_achieved = False
        self._max_iterations = max_iterations

        # attributes for journalling functionality, based on lstbx's
        # journaled_non_linear_ls class
        if tracking is None:
            # set default tracking
            tracking = refinery_phil_scope.extract().refinery.journal
        self.history = Journal()
        self.history.add_column("num_reflections")
        self.history.add_column("objective")  # flex.double()
        self.history.add_column("gradient_norm")  # flex.double()
        self.history.add_column("solution_norm")  # flex.double()
        self.history.add_column("parameter_vector")
        self.history.add_column("parameter_vector_norm")  # flex.double()
        self.history.add_column("rmsd")

        # Optional tracking
        if tracking.track_step:
            self.history.add_column("solution")
        if tracking.track_gradient:
            self.history.add_column("gradient")
        if tracking.track_parameter_correlation:
            self.history.add_column("parameter_correlation")
        if tracking.track_condition_number:
            self.history.add_column("condition_number")
        if tracking.track_jacobian_structure:
            self.history.add_column("jacobian_structure")
        if tracking.track_normal_matrix:
            self.history.add_column("normal_matrix")
        if tracking.track_out_of_sample_rmsd:
            self.history.add_column("out_of_sample_rmsd")

        # number of processes to use, for engines that support multiprocessing
        self._nproc = 1

        self.prepare_for_step()

    def get_num_steps(self):
        return self.history.get_nrows() - 1

    def prepare_for_step(self):
        """Update the parameterisation and prepare the target function"""

        x = self.x
        if self._constr_manager is not None:
            x = self._constr_manager.expand_parameters(x)

        # set current parameter values
        self._parameters.set_param_vals(x)

        # do reflection prediction
        self._target.predict()

    def update_journal(self):
        """Append latest step information to the journal attributes"""

        # add step quantities to journal
        self.history.add_row()
        self.history.set_last_cell("num_reflections", self._target.get_num_matches())
        self.history.set_last_cell("rmsd", self._target.rmsds())
        self.history.set_last_cell(
            "parameter_vector", self._parameters.get_param_vals()
        )
        self.history.set_last_cell("objective", self._f)
        if "gradient" in self.history:
            self.history.set_last_cell("gradient", self._g)
        if "parameter_correlation" in self.history and self._jacobian is not None:
            resid_names = [s.replace("RMSD_", "") for s in self._target.rmsd_names]
            # split Jacobian into dense matrix blocks corresponding to each residual
            jblocks = self.split_jacobian_into_blocks()
            corrmats = {}
            for r, j in zip(resid_names, jblocks):
                corrmats[r] = self._packed_corr_mat(j)
            self.history.set_last_cell("parameter_correlation", corrmats)
        if "jacobian_structure" in self.history and self._jacobian is not None:
            self.history.set_last_cell("jacobian_structure", self._jacobian_structure())
        if "condition_number" in self.history and self._jacobian is not None:
            self.history.set_last_cell(
                "condition_number", self.jacobian_condition_number()
            )
        if "out_of_sample_rmsd" in self.history:
            preds = self._target.predict_for_free_reflections()
            self.history.set_last_cell(
                "out_of_sample_rmsd", self._target.rmsds_for_reflection_table(preds)
            )

    def split_jacobian_into_blocks(self) -> List[flex.double]:
        """Split the Jacobian into blocks each corresponding to a separate
        residual, converting sparse to flex.double if appropriate"""

        nblocks = len(self._target.rmsd_names)

        try:
            # The Jacobian might be a sparse matrix
            j: flex.double = self._jacobian.as_dense_matrix()
        except AttributeError:
            j = self._jacobian

        nr, nc = j.all()
        nr_block = int(nr / nblocks)
        row_start = [e * nr_block for e in range(nblocks)]
        blocks = [j.matrix_copy_block(rs, 0, nr_block, nc) for rs in row_start]

        return blocks

    @staticmethod
    def _packed_corr_mat(m: flex.double) -> List[float]:
        """Return a list containing the upper diagonal values of the
        correlation matrix calculated between columns of 2D matrix flex.double
        matrix m"""

        _, nc = m.all()

        tmp = []
        for col1 in range(nc):
            for col2 in range(col1, nc):
                if col1 == col2:
                    tmp.append(1.0)
                else:
                    # Avoid spuriously high correlation between a column that should be
                    # zero (such as the gradient of X residuals wrt the Shift2 parameter)
                    # and another column (such as the gradient of X residuals wrt the
                    # Dist parameter) by rounding values to 15 places. It seems that such
                    # spurious correlations may occur in cases where gradients are
                    # calculated to be zero by matrix operations, rather than set to zero.
                    v1 = m.matrix_copy_column(col1).round(15)
                    v2 = m.matrix_copy_column(col2).round(15)
                    tmp.append(flex.linear_correlation(v1, v2).coefficient())

        return tmp

    def get_correlation_matrix_for_step(self, step):
        """For each type of residual (e.g. X, Y, Phi), decompress and return the
        full 2D correlation matrix between columns of the Jacobian that was
        stored in the journal at the given step number. If not available, return
        None"""

        if "parameter_correlation" not in self.history:
            return None
        try:
            packed_mats = self.history["parameter_correlation"][step]
        except IndexError:
            return None
        if packed_mats is None:
            return None

        packed_mats = copy.deepcopy(packed_mats)

        nparam = len(self._parameters)

        for k, v in packed_mats.items():
            corr_mat = flex.double(flex.grid(nparam, nparam))
            i = 0
            for row in range(nparam):
                for col in range(row, nparam):
                    corr_mat[row, col] = v[i]
                    i += 1
            corr_mat.matrix_copy_upper_to_lower_triangle_in_place()
            packed_mats[k] = corr_mat
        return packed_mats

    def jacobian_condition_number(self):
        """Calculate the condition number of the Jacobian, for tracking in the
        refinement journal, if requested. The condition number of a matrix A is
        defined as cond(A) = ||A|| ||inv(A)||. For a rectangular matrix the inverse
        operation refers to the Moore-Penrose pseudoinverse. Various matrix norms
        can be used, resulting in numerically different condition numbers, however
        the 2-norm is commonly used. In that case, the definition is equivalent
        to the ratio of the largest to smallest singular values of the matrix:
        cond(A) = sig_(A) / sig_min(A). That is the calculation that is performed
        here.

        The condition number is a measure of how accurate the solution x to the
        equation Ax = b will be. Essentially it measures how errors are amplified
        through the linear equation. The condition number is large in the case that
        the columns of A are nearly linearly-dependent (and infinite for a singular
        matrix). We use it here then to detect situations where the correlation
        between effects of different parameter shifts becomes large and therefore
        refinement is problematic.

        Note, the Jacobian used here does not include any additional rows due to
        restraints terms that might be applied, or any parameter reduction due to
        constraints. Therefore this condition number relates to the pure linearised
        (Gauss-Newton) step, which might not actually be what the refinement engine
        uses. It can be indicative of issues in the fundamental set up of the least
        squares problem, even if these issues are avoided in practice (e.g. by
        use of an algorithm like Levenberg-Marquardt, inclusion of restraints or
        parameter reduction).
        """
        try:
            # The Jacobian might be a sparse matrix
            j = self._jacobian.as_dense_matrix().deep_copy()
        except AttributeError:
            j = self._jacobian.deep_copy()

        from scitbx.linalg.svd import real as svd_real

        svd = svd_real(j, False, False)

        # The condition number is the ratio of the largest to the smallest singular
        # values of the matrix
        return max(svd.sigma) / min(svd.sigma)

    def test_for_termination(self):
        """Return True if refinement should be terminated"""

        # Basic version delegate to the Target class. Derived classes may
        # implement other termination criteria
        self._target_achieved = self._target.achieved()

        return self._target_achieved

    def test_rmsd_convergence(self):
        """Test for convergence of RMSDs"""

        # http://en.wikipedia.org/wiki/
        # Non-linear_least_squares#Convergence_criteria
        try:
            r1 = self.history["rmsd"][-1]
            r2 = self.history["rmsd"][-2]
        except IndexError:
            return False

        tests = [
            abs((e[1] - e[0]) / e[1]) < 0.0001 if e[1] > 0 else True
            for e in zip(r1, r2)
        ]

        return all(tests)

    def test_objective_increasing_but_not_nref(self):
        """Test for an increase in the objective value between steps. This
        could be caused simply by the number of matches between observations
        and predictions increasing. However, if the number of matches stayed
        the same or reduced then this is a bad sign."""

        try:
            l1 = self.history["objective"][-1]
            l2 = self.history["objective"][-2]
            n1 = self.history["num_reflections"][-1]
            n2 = self.history["num_reflections"][-2]
        except IndexError:
            return False

        return l1 > l2 and n1 <= n2

    def set_nproc(self, nproc):
        """Set number of processors for multiprocessing. Override in derived classes
        if a policy dictates that this must not be user-controlled"""
        self._nproc = nproc

    def run(self):
        """
        To be implemented by derived class. It is expected that each step of
        refinement be preceded by a call to prepare_for_step and followed by
        calls to update_journal and test_for_termination (in that order).
        """

        # Specify a minimizer and its parameters, and run
        raise NotImplementedError()

    def _dump_normal_matrix(self):
        """Output the full normal matrix at the current step as a string for debugging."""
        try:
            s = (
                self.normal_matrix_packed_u()
                .matrix_packed_u_as_symmetric()
                .as_scitbx_matrix()
                .matlab_form(format=None, one_row_per_line=True)
            )
        except AttributeError:
            s = "Unavailable"
        return s

    def _jacobian_structure(self):
        """Calculate the structure of the Jacobian matrix and output as a table for
        display"""
        result = []
        p_names = self._parameters.get_param_names()
        for name, col in zip(p_names, self._jacobian.cols()):
            result.append(
                {
                    "parameter": name,
                    "nrows": col.size,
                    "structural_zeroes": (col.size - col.non_zeroes),
                    "all_zeroes": ((col.as_dense_vector() == 0.0).count(True)),
                }
            )
        return result


class DisableMPmixin:
    """A mixin class that disables setting of nproc for multiprocessing"""

    def set_nproc(self, nproc):
        if nproc != 1:
            raise NotImplementedError()


class AdaptLbfgs(Refinery):
    """Adapt Refinery for L-BFGS minimiser"""

    def __init__(self, *args, **kwargs):

        Refinery.__init__(self, *args, **kwargs)

        self._termination_params = lbfgs.termination_parameters(
            max_iterations=self._max_iterations
        )

        self._log_string = StringIO

    def compute_functional_and_gradients(self):
        L, dL_dp, _ = self.compute_functional_gradients_and_curvatures()
        self._f = L
        self._g = dL_dp
        return self._f, self._g

    def compute_functional_gradients_and_curvatures(self):

        self.prepare_for_step()

        # observation terms
        blocks = self._target.split_matches_into_blocks(nproc=self._nproc)
        if self._nproc > 1:
            task_results = easy_mp.parallel_map(
                func=self._target.compute_functional_gradients_and_curvatures,
                iterable=blocks,
                processes=self._nproc,
                method="multiprocessing",
                preserve_exception_message=True,
            )

        else:
            task_results = [
                self._target.compute_functional_gradients_and_curvatures(block)
                for block in blocks
            ]

        # reduce blockwise results
        flist, glist, clist = zip(*task_results)
        f = sum(flist)
        g = [sum(g) for g in zip(*glist)]
        c = [sum(c) for c in zip(*clist)]

        # restraints terms
        restraints = (
            self._target.compute_restraints_functional_gradients_and_curvatures()
        )

        if restraints:
            f += restraints[0]
            g = [a + b for a, b in zip(g, restraints[1])]
            c = [a + b for a, b in zip(c, restraints[2])]

        # compact and reorder according to the constraints
        if self._constr_manager is not None:
            g = self._constr_manager.constrain_gradient_vector(g)
            c = self._constr_manager.constrain_gradient_vector(c)

        return f, flex.double(g), flex.double(c)

    def callback_after_step(self, minimizer):
        """
        Do journalling, evaluate rmsds and return True if the target is
        reached to terminate the refinement.
        """

        self.update_journal()
        logger.debug("Step %d", self.history.get_nrows() - 1)

        if self.test_for_termination():
            self.history.reason_for_termination = TARGET_ACHIEVED
            return True

        if self.test_rmsd_convergence():
            self.history.reason_for_termination = RMSD_CONVERGED
            return True

        return False

    def run_lbfgs(self, curvatures=False):
        """
        Run the minimiser, keeping track of its log.
        """

        ref_log = self._log_string()
        if curvatures:
            self.diag_mode = "always"
        self.minimizer = lbfgs.run(
            target_evaluator=self,
            termination_params=self._termination_params,
            log=ref_log,
        )

        log = ref_log.getvalue()
        if self._log:
            with open(self._log, "a") as f:
                f.write(log)
        ref_log.close()

        pos = log.rfind("lbfgs minimizer stop: ")
        if pos >= 0:
            msg = log[pos:].splitlines()[0]
            if self.history.reason_for_termination:
                self.history.reason_for_termination += "\n"
                self.history.reason_for_termination += msg
            else:
                self.history.reason_for_termination = msg

        if self.minimizer.error:
            self.history.reason_for_termination = self.minimizer.error


class SimpleLBFGS(AdaptLbfgs):
    """Refinery implementation, using cctbx LBFGS with basic settings"""

    def run(self):
        return self.run_lbfgs(curvatures=False)


class LBFGScurvs(AdaptLbfgs):
    """Refinery implementation using cctbx LBFGS with curvatures"""

    def run(self):
        return self.run_lbfgs(curvatures=True)

    def compute_functional_gradients_diag(self):

        L, dL_dp, curvs = self.compute_functional_gradients_and_curvatures()
        self._f = L
        self._g = dL_dp

        # Curvatures of zero will cause a crash, because their inverse is taken.
        assert curvs.all_gt(0.0)

        diags = 1.0 / curvs

        msg = "  curv: " + "%.5f " * len(tuple(curvs))
        logger.debug(msg, *curvs)

        return self._f, self._g, diags


class AdaptLstbx(Refinery, normal_eqns.non_linear_ls, normal_eqns.non_linear_ls_mixin):
    """Adapt Refinery for lstbx"""

    def __init__(
        self,
        target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=None,
    ):

        Refinery.__init__(
            self,
            target,
            prediction_parameterisation,
            constraints_manager,
            log=log,
            tracking=tracking,
            max_iterations=max_iterations,
        )

        # required for restart to work (do I need that method?)
        self.x_0 = self.x.deep_copy()

        # keep attribute for the Cholesky factor required for ESD calculation
        self.cf = None

        normal_eqns.non_linear_ls.__init__(self, n_parameters=len(self.x))

    def restart(self):
        self.x = self.x_0.deep_copy()
        self.old_x = None

    def parameter_vector_norm(self):
        return self.x.norm()

    def build_up(self, objective_only=False):

        # code here to calculate the residuals. Rely on the target class
        # for this

        # I need to use the weights. They are the variances of the
        # observations... See http://en.wikipedia.org/wiki/Non-linear_least_squares
        # at 'diagonal weight matrix'

        # set current parameter values
        self.prepare_for_step()

        # Reset the state to construction time, i.e. no equations accumulated
        self.reset()

        # observation terms
        if objective_only:
            residuals, weights = self._target.compute_residuals()
            self.add_residuals(residuals, weights)
        else:
            blocks = self._target.split_matches_into_blocks(nproc=self._nproc)

            if self._nproc > 1:

                # ensure the jacobian is not tracked
                self._jacobian = None

                # processing functions
                def task_wrapper(block):
                    (
                        residuals,
                        jacobian,
                        weights,
                    ) = self._target.compute_residuals_and_gradients(block)
                    return {
                        "residuals": residuals,
                        "jacobian": jacobian,
                        "weights": weights,
                    }

                def callback_wrapper(result):
                    j = result["jacobian"]
                    if self._constr_manager is not None:
                        j = self._constr_manager.constrain_jacobian(j)
                    self.add_equations(result["residuals"], j, result["weights"])
                    # no longer need the result
                    result["residuals"] = None
                    result["jacobian"] = None
                    result["weights"] = None
                    return

                easy_mp.parallel_map(
                    func=task_wrapper,
                    iterable=blocks,
                    processes=self._nproc,
                    callback=callback_wrapper,
                    method="multiprocessing",
                    preserve_exception_message=True,
                )

            else:
                for block in blocks:
                    (
                        residuals,
                        jacobian,
                        weights,
                    ) = self._target.compute_residuals_and_gradients(block)
                    if self._constr_manager is not None:
                        jacobian = self._constr_manager.constrain_jacobian(jacobian)

                    self.add_equations(residuals, jacobian, weights)

                # Keep reference to the Jacobian in case required by the Journal
                self._jacobian = jacobian

        # restraints terms
        restraints = self._target.compute_restraints_residuals_and_gradients()
        if restraints:
            if objective_only:
                self.add_residuals(restraints[0], restraints[2])
            else:
                j = restraints[1]
                if self._constr_manager is not None:
                    j = self._constr_manager.constrain_jacobian(j)
                self.add_equations(restraints[0], j, restraints[2])

    def step_forward(self):
        self.old_x = self.x.deep_copy()
        self.x += self.step()

    def step_backward(self):
        if self.old_x is None:
            return False
        else:
            self.x, self.old_x = self.old_x, None
            return True

    def set_cholesky_factor(self):
        """Set the Cholesky factor required for ESD calculation. This method is
        valid only for the LSTBX dense matrix interface"""

        self.cf = self.step_equations().cholesky_factor_packed_u().deep_copy()

    def calculate_esds(self):
        """Calculate ESDs of parameters"""

        # it is possible to get here with zero steps taken by the minimiser. For
        # example by failing for the MAX_TRIAL_ITERATIONS reason before any forward
        # steps are taken with the LevMar engine. If so the below is invalid,
        # so return early
        if self.history.get_nrows() == 0:
            return None

        if self.cf is None:
            return None

        # if constraints were used then the normal matrix has fewer rows/columns
        # than the number of expanded parameters. At the moment, do not support
        # this calculation when constraints were used
        if self._constr_manager is not None:
            return None

        # invert normal matrix from N^-1 = (U^-1)(U^-1)^T
        cf_inv = self.cf.matrix_packed_u_as_upper_triangle().matrix_inversion()
        nm_inv = cf_inv.matrix_multiply_transpose(cf_inv)

        # keep the estimated parameter variance-covariance matrix
        self.parameter_var_cov = self.history["reduced_chi_squared"][-1] * nm_inv
        # send this back to the models to calculate their uncertainties
        self._parameters.calculate_model_state_uncertainties(self.parameter_var_cov)

        # send parameter variances back to the parameter classes
        # themselves, for reporting purposes and for building restraints
        # based on existing parameterisations.
        s2 = self.parameter_var_cov.matrix_diagonal()
        assert s2.all_ge(0.0)
        s = flex.sqrt(s2)
        self._parameters.set_param_esds(s)


class GaussNewtonIterations(AdaptLstbx, normal_eqns_solving.iterations):
    """Refinery implementation, using lstbx Gauss Newton iterations"""

    # defaults that may be overridden
    gradient_threshold = 1.0e-10
    step_threshold = None
    damping_value = 0.0007
    max_shift_over_esd = 15
    convergence_as_shift_over_esd = 1e-5

    def __init__(
        self,
        target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=20,
        **kwds,
    ):

        AdaptLstbx.__init__(
            self,
            target,
            prediction_parameterisation,
            constraints_manager,
            log=log,
            tracking=tracking,
            max_iterations=max_iterations,
        )

        # add an attribute to the journal
        self.history.add_column("reduced_chi_squared")  # flex.double()

        # adopt any overrides of the defaults above
        libtbx.adopt_optional_init_args(self, kwds)

    def run(self):
        self.n_iterations = 0

        # prepare for first step
        self.build_up()

        # return early if refinement is not possible
        if self.dof < 1:
            self.history.reason_for_termination = DOF_TOO_LOW
            return

        while True:

            # set functional and gradients for the step (to add to the history)
            self._f = self.objective()
            self._g = -self.opposite_of_gradient()

            # cache some items for the journal prior to solve
            pvn = self.parameter_vector_norm()
            gn = self.opposite_of_gradient().norm_inf()
            if "normal_matrix" in self.history:
                nm = self._dump_normal_matrix()

            # solve the normal equations
            self.solve()

            # standard journalling
            self.update_journal()
            logger.debug("Step %d", self.history.get_nrows() - 1)

            # add cached items to the journal
            self.history.set_last_cell("parameter_vector_norm", pvn)
            self.history.set_last_cell("gradient_norm", gn)
            if "normal_matrix" in self.history:
                self.history.set_last_cell("normal_matrix", nm)

            # extra journalling post solve
            if "solution" in self.history:
                self.history.set_last_cell("solution", self.actual.step().deep_copy())
            self.history.set_last_cell("solution_norm", self.step().norm())
            self.history.set_last_cell("reduced_chi_squared", self.chi_sq())

            # test termination criteria
            if self.test_for_termination():
                self.history.reason_for_termination = TARGET_ACHIEVED
                break

            if self.test_rmsd_convergence():
                self.history.reason_for_termination = RMSD_CONVERGED
                break

            if self.had_too_small_a_step():
                self.history.reason_for_termination = STEP_TOO_SMALL
                break

            if self.test_objective_increasing_but_not_nref():
                self.history.reason_for_termination = OBJECTIVE_INCREASE
                if self.step_backward():
                    self.history.reason_for_termination += (
                        ". Parameters set back one step"
                    )
                self.prepare_for_step()
                break

            if self.n_iterations == self._max_iterations:
                self.history.reason_for_termination = MAX_ITERATIONS
                break

            # prepare for next step
            self.step_forward()
            self.n_iterations += 1
            self.build_up()

        self.set_cholesky_factor()
        self.calculate_esds()


class LevenbergMarquardtIterations(GaussNewtonIterations):
    """Refinery implementation, employing lstbx Levenberg Marquadt
    iterations"""

    tau = 1e-3

    @property
    def mu(self):
        return self._mu

    @mu.setter
    def mu(self, value):
        self._mu = value

    def setup_mu(self):
        """Setup initial value for mu"""
        a = self.normal_matrix_packed_u()
        self.mu = self.tau * flex.max(a.matrix_packed_u_diagonal())

    def add_constant_to_diagonal(self, mu):
        """Add the constant value mu to the diagonal of the normal matrix"""
        a = self.normal_matrix_packed_u()
        a.matrix_packed_u_diagonal_add_in_place(self.mu)

    def report_progress(self, objective):
        """Callback within the refinement main loop that can be overridden to
        report the value of the objective function (and possibly) other details for
        long-running methods"""
        pass

    def _run_core(self):

        # add an attribute to the journal
        self.history.add_column("mu")
        self.history.add_column("nu")

        # set max iterations if not already.
        if self._max_iterations is None:
            self._max_iterations = 100

        self.n_iterations = 0
        nu = 2
        self.build_up()

        # early test for linear independence, require all right hand side elements to be non-zero
        RHS = self.step_equations().right_hand_side()
        if RHS.count(0.0) > 0:
            p_names = [
                b for a, b in zip(RHS, self._parameters.get_param_names()) if a == 0.0
            ]
            raise DialsRefineRuntimeError(
                f"The normal equations have an indeterminate solution. The problematic parameters are {', '.join(p_names)}."
            )

        # return early if refinement is not possible
        if self.dof < 1:
            self.history.reason_for_termination = DOF_TOO_LOW
            return

        self.setup_mu()

        while True:

            # set functional and gradients for the step
            self._f = self.objective()
            self._g = -self.opposite_of_gradient()

            # cache some items for the journal prior to solve
            pvn = self.parameter_vector_norm()
            gn = self.opposite_of_gradient().norm_inf()
            if "normal_matrix" in self.history:
                nm = self._dump_normal_matrix()

            self.add_constant_to_diagonal(self.mu)

            # solve the normal equations
            self.solve()

            # keep the cholesky factor for ESD calculation if we end this step. Doing
            # it here ensures the normal equations are solved (cholesky_factor_packed_u
            # can only be called if that is the case)
            self.set_cholesky_factor()

            # standard journalling
            self.update_journal()
            logger.debug("Step %d", self.history.get_nrows() - 1)

            # add cached items to the journal
            self.history.set_last_cell("parameter_vector_norm", pvn)
            self.history.set_last_cell("gradient_norm", gn)
            if "normal_matrix" in self.history:
                self.history.set_last_cell("normal_matrix", nm)

            # extra journalling post solve
            self.history.set_last_cell("mu", self.mu)
            self.history.set_last_cell("nu", nu)
            if "solution" in self.history:
                self.history.set_last_cell("solution", self.actual.step().deep_copy())
            self.history.set_last_cell("solution_norm", self.step().norm())
            self.history.set_last_cell("reduced_chi_squared", self.chi_sq())

            # test termination criteria before taking the next forward step
            if self.had_too_small_a_step():
                self.history.reason_for_termination = STEP_TOO_SMALL
                break
            if self.test_for_termination():
                self.history.reason_for_termination = TARGET_ACHIEVED
                break
            if self.test_rmsd_convergence():
                self.history.reason_for_termination = RMSD_CONVERGED
                break
            if self.n_iterations == self._max_iterations:
                self.history.reason_for_termination = MAX_ITERATIONS
                break

            h = self.step()
            expected_decrease = 0.5 * h.dot(self.mu * h - self._g)
            if expected_decrease == 0:
                self.history.reason_for_termination = OBJECTIVE_FLAT
                break
            self.step_forward()
            self.n_iterations += 1
            self.build_up(objective_only=True)
            objective_new = self.objective()
            self.report_progress(objective_new)
            actual_decrease = self._f - objective_new
            rho = actual_decrease / expected_decrease
            if rho > 0:
                self.mu *= max(1 / 3, 1 - (2 * rho - 1) ** 3)
                nu = 2
            else:
                self.step_backward()
                self.history.del_last_row()
                if nu >= 8192:
                    self.history.reason_for_termination = MAX_TRIAL_ITERATIONS
                    break
                self.mu *= nu
                nu *= 2

            # prepare for next step
            self.build_up()

        self.calculate_esds()

    def run(self):
        self._run_core()
        self.calculate_esds()



from __future__ import annotations

import logging
from math import floor, sqrt

import numpy as np
from numpy.linalg import inv, norm

from cctbx.array_family import flex
from dxtbx import flumpy
from scitbx import matrix

from dials.algorithms.profile_model.ellipsoid import chisq_quantile
from dials.algorithms.statistics.fast_mcd import FastMCD, maha_dist_sq

logger = logging.getLogger("dials")


def _index(reflection_table, experiment, fail_on_bad_index=False):
    """Index the strong spots"""

    # Get some stuff from experiment
    A = np.array(experiment.crystal.get_A(), dtype=np.float64).reshape(3, 3)
    s0 = np.array([experiment.beam.get_s0()], dtype=np.float64).reshape(3, 1)
    s0_length = norm(s0)
    detector = experiment.detector

    # Create array if necessary
    if "miller_index" not in reflection_table:
        reflection_table["miller_index"] = flex.miller_index(len(reflection_table))

    # Index all the reflections
    miller_index = reflection_table["miller_index"]
    selection = flex.size_t()
    num_reindexed = 0
    for i, xyz in enumerate(reflection_table["xyzobs.px.value"]):
        # Get the observed pixel coordinate
        x, y, _ = xyz

        # Get the lab coord
        s1 = np.array(
            detector[0].get_pixel_lab_coord((x, y)), dtype=np.float64
        ).reshape(3, 1)
        s1_norm = norm(s1)
        s1 *= s0_length / s1_norm

        # Get the reciprocal lattice vector
        r = s1 - s0
        # Compute the fractional miller index
        hf = np.matmul(inv(A), r)
        # Compute the integer miller index
        h = np.array([int(floor(j + 0.5)) for j in hf[:, 0]], dtype=int).reshape(3, 1)

        # Print warning if reindexing
        if tuple(h) != miller_index[i]:
            logger.warn(
                "Reindexing (% 3d, % 3d, % 3d) -> (% 3d, % 3d, % 3d)"
                % (miller_index[i] + tuple(h))
            )
            num_reindexed += 1
            miller_index[i] = matrix.col(flumpy.from_numpy(h))
            if fail_on_bad_index:
                raise RuntimeError("Bad index")

        # If its not indexed as 0, 0, 0 then append
        if h.any() and norm(h - hf) < 0.3:
            selection.append(i)

    # Print some info
    logger.info(
        "Reindexed %d/%d input reflections" % (num_reindexed, len(reflection_table))
    )
    logger.info(
        "Selected %d/%d input reflections" % (len(selection), len(reflection_table))
    )

    # Select all the indexed reflections
    reflection_table.set_flags(selection, reflection_table.flags.indexed)
    reflection_table = reflection_table.select(selection)
    return reflection_table


def _predict(reflection_table, experiment):
    """
    Predict the position of the spots

    """

    # Get some stuff from experiment
    A = np.array(experiment.crystal.get_A(), dtype=np.float64).reshape((3, 3))
    s0 = np.array([experiment.beam.get_s0()], dtype=np.float64).reshape(3, 1)
    s0_length = norm(s0)

    # Compute the vector to the reciprocal lattice point
    # since this is not on the ewald sphere, lets call it s2
    s1 = flex.vec3_double(reflection_table.size())
    s2 = flex.vec3_double(reflection_table.size())
    for i, h in enumerate(reflection_table["miller_index"]):
        r = np.matmul(A, np.array([h], dtype=np.float64).reshape(3, 1))
        s2_i = r + s0
        s2[i] = matrix.col(flumpy.from_numpy(s2_i))
        s1[i] = matrix.col(flumpy.from_numpy(s2_i * s0_length / norm(s2_i)))
    reflection_table["s1"] = s1
    reflection_table["s2"] = s2
    reflection_table["entering"] = flex.bool(reflection_table.size(), False)

    # Compute the ray intersections
    xyzpx = flex.vec3_double()
    xyzmm = flex.vec3_double()
    for ss in s1:
        mm = experiment.detector[0].get_ray_intersection(ss)
        px = experiment.detector[0].millimeter_to_pixel(mm)
        xyzpx.append(px + (0,))
        xyzmm.append(mm + (0,))
    reflection_table["xyzcal.mm"] = xyzmm
    reflection_table["xyzcal.px"] = xyzpx
    logger.info("Do prediction for %d reflections" % len(reflection_table))
    return reflection_table


def _filter_reflections_based_on_centroid_distance(
    reflection_table,
    experiment,
    outlier_probability=0.975,
    max_separation=2,
):
    """
    Filter reflections too far from predicted position

    """

    # Compute the x and y residuals
    Xobs, Yobs, _ = reflection_table["xyzobs.px.value"].parts()
    Xcal, Ycal, _ = reflection_table["xyzcal.px"].parts()
    Xres = Xobs - Xcal
    Yres = Yobs - Ycal

    # Compute the epsilon residual
    s0_length = 1.0 / experiment.beam.get_wavelength()
    s1x, s1y, s1z = reflection_table["s2"].parts()
    s1_length = flex.sqrt(s1x**2 + s1y**2 + s1z**2)
    Eres = s1_length - s0_length

    # Initialise the fast_mcd outlier algorithm
    # fast_mcd = FastMCD((Xres, Yres, Eres))
    try:
        fast_mcd = FastMCD((Xres, Yres))
    except AssertionError as e:
        raise RuntimeError(e)

    # get location and MCD scatter estimate
    T, S = fast_mcd.get_corrected_T_and_S()

    # get squared Mahalanobis distances
    # d2s = maha_dist_sq((Xres, Yres, Eres), T, S)
    d2s = maha_dist_sq((Xres, Yres), T, S)

    # Compute the cutoff
    mahasq_cutoff = chisq_quantile(2, outlier_probability)

    # compare to the threshold and select reflections
    selection1 = d2s < mahasq_cutoff
    selection2 = flex.sqrt(Xres**2 + Yres**2) < max_separation
    selection = selection1 & selection2
    reflection_table = reflection_table.select(selection)
    n_refl = reflection_table.size()

    # Print some stuff
    logger.info("-" * 80)
    logger.info("Centroid outlier rejection")
    logger.info(f" Using MCD algorithm with probability = {outlier_probability}")
    logger.info(" Max X residual: %f" % flex.max(flex.abs(Xres)))
    logger.info(" Max Y residual: %f" % flex.max(flex.abs(Yres)))
    logger.info(" Max E residual: %f" % flex.max(flex.abs(Eres)))
    logger.info(" Mean X RMSD: %f" % (sqrt(flex.sum(Xres**2) / len(Xres))))
    logger.info(" Mean Y RMSD: %f" % (sqrt(flex.sum(Yres**2) / len(Yres))))
    logger.info(" Mean E RMSD: %f" % (sqrt(flex.sum(Eres**2) / len(Eres))))
    logger.info(" MCD location estimate: %.4f, %.4f" % tuple(T))
    logger.info(
        """ MCD scatter estimate:
    %.7f, %.7f,
    %.7f, %.7f"""
        % tuple(S)
    )
    logger.info(" Number of outliers: %d" % selection1.count(False))
    logger.info(
        " Number of reflections with residual > %0.2f pixels: %d"
        % (max_separation, selection2.count(False))
    )
    logger.info(f"Number of reflections selection for refinement: {n_refl}")
    logger.info("-" * 80)

    return reflection_table


def reindex(
    reflection_table,
    experiment,
    outlier_probability=0.975,
    max_separation=2,
    fail_on_bad_index=False,
):
    """Reindex strong spots and perform filtering"""
    reflection_table = _index(reflection_table, experiment, fail_on_bad_index)
    reflection_table = _predict(reflection_table, experiment)
    reflection_table = _filter_reflections_based_on_centroid_distance(
        reflection_table,
        experiment,
        outlier_probability=outlier_probability,
        max_separation=max_separation,
    )
    return reflection_table



"""
Classes that define a component of a scaling model.

Each class holds the parameters and relevant data, as a list of
arrays, from which to calculate inverse scale factors and
derivatives.
The components are initialised without any data, which is added
by setting the data dict. In order to update the internal data
lists in order to calculate the scales and derivatives, the
update_reflection_data method should be called, which can optionally
be provided with selection arrays to split the data for blockwise/parallel
calculations.

The scaling algorithm makes use of the components in the following way.
First, the data for all 'suitable' reflections are added to the components.
Then, at different stages of the algorithm, selection lists are provided
to select a subset of this data (e.g. a small subset to prepare the
component for minimisation calculation, or a large subset for calculating
the scales for all reflections). The selection lists typically come from
the Ih_table datastructures so that the data in the components is split in
the same way as the data in the Ih_table datastructure.
"""

from __future__ import annotations

from scitbx import sparse

from dials.array_family import flex
from dials_scaling_ext import calculate_harmonic_tables_from_selections


class ScaleComponentBase:
    """
    Base scale component class.

    This defines an interface to access the parameters, the component
    of the inverse scale factor and it's derivatives with respect to
    the parameters. Scale components derived from the base class are
    designed to be instantiated by a ScalingModel class, by supplying
    an initial array of parameters and optionally the current estimated
    standard deviations. The relevant data from a reflection table is
    added later by a Scaler using the update_reflection_data method.
    This behaviour allows data to easily be added/changed after selecting
    subsets of the data.
    """

    def __init__(self, initial_values, parameter_esds=None):
        """Set the initial parameter values, parameter esds and n_params."""
        self._parameters = initial_values
        self._parameter_esds = parameter_esds
        self._n_params = len(self._parameters)
        self._var_cov = None
        self._n_refl = []  # store as a list, to allow holding of data in blocks
        self._parameter_restraints = None
        self._data = {}

    @property
    def data(self):
        """
        Return a dictionary of reflection data relevant to the particular component.

        This is designed to be a dict of arrays which can be selected from when
        updating the component (i.e. selecting subsets).
        """
        return self._data

    @data.setter
    def data(self, data):
        self._data = data

    @data.deleter
    def data(self):
        self._data = {}

    @property
    def parameter_restraints(self):
        """Restraint weights for the component parameters."""
        return self._parameter_restraints

    @parameter_restraints.setter
    def parameter_restraints(self, restraints):
        assert restraints.size() == self.parameters.size()
        self._parameter_restraints = restraints

    @property
    def n_params(self):
        """Get the number of parameters of the component (read-only)."""
        return self._n_params

    @property
    def parameters(self):
        """Parameters of the component."""
        return self._parameters

    @parameters.setter
    def parameters(self, new_parameters):
        assert len(new_parameters) == len(
            self._parameters
        ), f"""
attempting to set a new set of parameters of different length than previous
assignment: was {len(self._parameters)}, attempting {len(new_parameters)}"""
        self._parameters = new_parameters

    @property
    def free_parameters(self):
        return self._parameters

    @free_parameters.setter
    def free_parameters(self, parameters):
        self._parameters = parameters

    @property
    def parameter_esds(self):
        """Return the estimated standard deviations of the parameters."""
        return self._parameter_esds

    @parameter_esds.setter
    def parameter_esds(self, esds):
        assert len(esds) == len(self.parameters)
        self._parameter_esds = esds

    @property
    def free_parameter_esds(self):
        """Return the estimated standard deviations of the parameters."""
        return self._parameter_esds

    @free_parameter_esds.setter
    def free_parameter_esds(self, esds):
        assert len(esds) == len(self.free_parameters)
        self._parameter_esds = esds

    def calculate_restraints(self):
        """Calculate residual and gradient restraints for the component."""
        return None

    def calculate_jacobian_restraints(self):
        """Calculate residual and jacobian restraints for the component."""
        return None

    @property
    def var_cov_matrix(self):
        """Return the variance-covariance matrix of the parameters."""
        return self._var_cov

    @var_cov_matrix.setter
    def var_cov_matrix(self, var_cov):
        self._var_cov = var_cov

    @property
    def n_refl(self):
        """Return a list of the number of reflections in each block."""
        return self._n_refl

    def update_reflection_data(self, selection=None, block_selections=None):
        """
        Update the internal data arrays.

        Use the data stored in self.data, optionally with a selection array
        or list of selections, to populate a list of internal arrays e.g n_refl,
        normalised_values etc. to allow scale and derivative calculations. If no
        selection arrays are provided, the internal arrays will be lists
        containing one array/value, depending on the data type needed for
        derivative and scale calculation.

        Args:
            selection: A flex.bool selection array to select a subset of the
                internal data.
            block_selections (list): A list of flex.size_t arrays to select
                subsets of the internal data.
        """
        raise NotImplementedError()

    def calculate_scales_and_derivatives(self, block_id=0):
        """Calculate and return inverse scales and derivatives for a given block."""
        raise NotImplementedError()

    def calculate_scales(self, block_id=0):
        """Calculate and return inverse scales for a given block."""
        raise NotImplementedError()


class SingleScaleFactor(ScaleComponentBase):
    """
    A model component consisting of a single global scale parameter.

    The inverse scale factor for every reflection is the parameter
    value itself and the derivatives are therefore all 1.0.
    """

    null_parameter_value = 1.0

    def __init__(self, initial_values, parameter_esds=None):
        """Set the initial parameter values, parameter esds and n_params."""
        assert (
            len(initial_values) == 1
        ), """
This model component can only hold a single parameter."""
        super().__init__(initial_values, parameter_esds)

    @ScaleComponentBase.data.setter
    def data(self, data):
        """Set the data dict in the parent class."""
        assert set(data.keys()) == {"id"}, set(data.keys())
        self._data = data

    def update_reflection_data(self, selection=None, block_selections=None):
        """
        Update the internal n_refl list.

        Use the data stored in self.data, optionally with a boolean selection array
        or list of flex.size_t index selections, to make a list of n_refl (of length
        1 or len(block_selections)) by inspecting the size of the selection result,
        in order to allow scale and derivative calculations.

        Args:
            selection: Optional, a flex.bool selection array to select a subset of
                the internal data.
            block_selections (list): Optional, a list of flex.size_t arrays to
                select subsets of the internal data.
        """
        data = self.data["id"]
        if selection:
            self._n_refl = [data.select(selection).size()]
        elif block_selections:
            self._n_refl = [data.select(sel).size() for sel in block_selections]
        else:
            self._n_refl = [data.size()]

    def calculate_scales_and_derivatives(self, block_id=0):
        """Calculate and return inverse scales and derivatives for a given block."""
        scales = flex.double(self.n_refl[block_id], self._parameters[0])
        derivatives = sparse.matrix(self.n_refl[block_id], 1)
        for i in range(self.n_refl[block_id]):
            derivatives[i, 0] = 1.0
        return scales, derivatives

    def calculate_scales(self, block_id=0):
        """Calculate and return inverse scales for a given block."""
        return flex.double(self.n_refl[block_id], self._parameters[0])


class SingleBScaleFactor(ScaleComponentBase):
    """
    A model component for a single global B-factor parameter.

    The inverse scale factor for each reflection is given by
    S = exp(B/(2 * d^2)), the derivatives are S/(2 * d^2).
    """

    null_parameter_value = 0.0

    def __init__(self, initial_values, parameter_esds=None):
        """Set the initial parameter values, parameter esds and n_params."""
        super().__init__(initial_values, parameter_esds)
        self._d_values = []

    @property
    def d_values(self):
        """Return a list of arrays of d-values associated with this component."""
        return self._d_values

    @ScaleComponentBase.data.setter
    def data(self, data):
        """Set the data dict in the parent class."""
        assert set(data.keys()) == {"d"}, set(data.keys())
        self._data = data

    def update_reflection_data(self, selection=None, block_selections=None):
        """
        Update the internal n_refl and d_values lists.

        Use the data stored in self.data, optionally with a boolean selection array
        or list of flex.size_t index selections, to make a lists of n_refl and
        d_value arrays (of length 1 or len(block_selections)), in order to allow
        scale and derivative calculations.

        Args:
            selection: Optional, a flex.bool selection array to select a subset of
                the internal data.
            block_selections (list): Optional, a list of flex.size_t arrays to
                select subsets of the internal data.
        """
        data = self.data["d"]
        if selection:
            self._d_values = [data.select(selection)]
        elif block_selections:
            self._d_values = [data.select(sel) for sel in block_selections]
        else:
            self._d_values = [data]
        self._n_refl = [dvalues.size() for dvalues in self._d_values]

    def calculate_scales_and_derivatives(self, block_id=0):
        """Calculate and return inverse scales and derivatives for a given block."""
        d_squared = self._d_values[block_id] * self._d_values[block_id]
        scales = flex.exp(
            flex.double(self._n_refl[block_id], self._parameters[0]) / (2.0 * d_squared)
        )
        derivatives = sparse.matrix(self._n_refl[block_id], 1)
        for i in range(self._n_refl[block_id]):
            derivatives[i, 0] = scales[i] / (2.0 * d_squared[i])
        return scales, derivatives

    def calculate_scales(self, block_id=0):
        """Calculate and return inverse scales for a given block."""
        scales = flex.exp(
            flex.double(self._n_refl[block_id], self._parameters[0])
            / (2.0 * (self._d_values[block_id] * self._d_values[block_id]))
        )
        return scales


class LinearDoseDecay(ScaleComponentBase):
    """
    A model component for a decay that depends linearly on dose
    (see Holton Acta D 2019 D75 113-122)

    For the dose dependent, the form is I = I0 exp(-ln(2) D/ Hd).
    Parameterise this as linear function of rotation with an overall factor
    to refine. T(r) = exp(Cr/d - i.e. a one parameter model with the overall
    'dose' proportional factor C.
    """

    null_parameter_value = 0.0

    def __init__(self, initial_values, parameter_esds=None):
        """Set the initial parameter values, parameter esds and n_params."""
        super().__init__(initial_values, parameter_esds)
        self._d_values = []
        self._x = []  # rotation/time

    @property
    def d_values(self):
        """Return a list of arrays of d-values associated with this component."""
        return self._d_values

    @ScaleComponentBase.data.setter
    def data(self, data):
        """Set the data dict in the parent class."""
        assert set(data.keys()) == {"d", "x"}, set(data.keys())
        self._data = data

    def update_reflection_data(self, selection=None, block_selections=None):
        """
        Update the internal n_refl and d_values lists.

        Use the data stored in self.data, optionally with a boolean selection array
        or list of flex.size_t index selections, to make a lists of n_refl and
        d_value arrays (of length 1 or len(block_selections)), in order to allow
        scale and derivative calculations.

        Args:
            selection: Optional, a flex.bool selection array to select a subset of
                the internal data.
            block_selections (list): Optional, a list of flex.size_t arrays to
                select subsets of the internal data.
        """
        d = self.data["d"]
        x = self.data["x"]
        if selection:
            self._d_values = [d.select(selection)]
            self._x = [x.select(selection)]
        elif block_selections:
            self._d_values = [d.select(sel) for sel in block_selections]
            self._x = [x.select(sel) for sel in block_selections]
        else:
            self._d_values = [d]
            self._x = [x]
        self._n_refl = [dvalues.size() for dvalues in self._d_values]

    def calculate_scales_and_derivatives(self, block_id=0):
        """Calculate and return inverse scales and derivatives for a given block."""
        scales = flex.exp(
            self._parameters[0] * self._x[block_id] / self._d_values[block_id]
        )
        derivatives = sparse.matrix(self._n_refl[block_id], 1)
        for i in range(self._n_refl[block_id]):
            derivatives[i, 0] = scales[i] * (
                self._x[block_id][i] / self._d_values[block_id][i]
            )
        return scales, derivatives

    def calculate_scales(self, block_id=0):
        """Calculate and return inverse scales for a given block."""
        scales = flex.exp(
            self._parameters[0] * self._x[block_id] / self._d_values[block_id]
        )
        return scales


class QuadraticDoseDecay(LinearDoseDecay):
    """
    A model component for a decay that depends quadratically on dose

    For the dose dependent, the form is I = I0 exp(-C D/ d^2).
    Parameterise this as linear function of rotation with an overall factor
    to refine. T(r) = exp(Cr/d^2) - i.e. a one parameter model with the overall
    'dose' proportional factor C.
    """

    def calculate_scales_and_derivatives(self, block_id=0):
        """Calculate and return inverse scales and derivatives for a given block."""
        scales = flex.exp(
            self._parameters[0] * self._x[block_id] / (self._d_values[block_id] ** 2)
        )
        derivatives = sparse.matrix(self._n_refl[block_id], 1)
        for i in range(self._n_refl[block_id]):
            derivatives[i, 0] = scales[i] * (
                self._x[block_id][i] / (self._d_values[block_id][i] ** 2)
            )
        return scales, derivatives

    def calculate_scales(self, block_id=0):
        """Calculate and return inverse scales for a given block."""
        scales = flex.exp(
            self._parameters[0] * self._x[block_id] / (self._d_values[block_id] ** 2)
        )
        return scales


class SHScaleComponent(ScaleComponentBase):
    """
    A model component for a spherical harmonic absorption correction.

    This component uses a set of spherical harmonic functions to define
    an absorption surface for the crystal. A matrix of spherical harmonic
    coefficients for the data is stored in self._harmonic_values and is
    used to calculate the scales and derivatives.
    The scale is given by S = 1 + (sum_l sum_m Clm * Ylm) where Clm are
    the model parameters and Ylm are the spherical harmonic coefficients,
    the derivatives are then simply the coefficients Ylm.
    """

    null_parameter_value = 0.0
    coefficients_list = None  # shared class variable to reduce memory load

    def __init__(self, initial_values, parameter_esds=None):
        """Set the initial parameter values, parameter esds and n_params."""
        super().__init__(initial_values, parameter_esds)
        self._harmonic_values = []
        self._matrices = []

    @property
    def harmonic_values(self):
        """Return the matrix of harmonic coefficients for the internal data."""
        return self._harmonic_values

    @property
    def sph_harm_table(self):
        """Return the matrix of the full harmonic coefficient for a reflection table."""
        return self._data["sph_harm_table"]

    @sph_harm_table.setter
    def sph_harm_table(self, sht):
        self._data["sph_harm_table"] = sht

    @ScaleComponentBase.data.setter
    def data(self, data):
        """Set the data dict in the parent class."""
        try:
            assert set(data.keys()) == {"s1_lookup", "s0_lookup"}, set(data.keys())
            self._mode = "memory"
        except AssertionError:
            assert set(data.keys()) == {"sph_harm_table"}, set(data.keys())
            self._mode = "speed"  # Note: only speedier for small datasets
        self._data = data

    def calculate_restraints(self):
        """Calculate residual and gradient restraints for the component."""
        residual = self.parameter_restraints * self._parameters * self._parameters
        gradient = 2.0 * self.parameter_restraints * self._parameters
        return residual, gradient

    def calculate_jacobian_restraints(self):
        """Calculate residual and jacobian restraints for the component."""
        jacobian = sparse.matrix(self.n_params, self.n_params)
        for i in range(self.n_params):
            jacobian[i, i] = 1.0
        return self._parameters, jacobian, self._parameter_restraints

    def update_reflection_data(self, selection=None, block_selections=None):
        """
        Update the internal n_refl and harmonic_values lists.

        Use the harmonic values matrix stored in self.data, optionally with a
        boolean selection array or list of flex.size_t index selections, to make
        lists of n_refl and harmonic_value arrays (of length 1 or
        len(block_selections)), in order to allow scale and derivative calculations.

        Args:
            selection: Optional, a flex.bool selection array to select a subset of
                the internal data.
            block_selections (list): Optional, a list of flex.size_t arrays to
                select subsets of the internal data.
        """
        if self._mode == "speed":
            self._update_reflection_data_speedmode(selection, block_selections)
        elif self._mode == "memory":
            self._update_reflection_data_memorymode(selection, block_selections)
        else:
            raise ValueError

    def _update_reflection_data_memorymode(self, selection=None, block_selections=None):
        if len(self.coefficients_list) != self.n_params:
            self.coefficients_list = self.coefficients_list[0 : self.n_params]
            # modify only for this instance, only needs to be done once per instance.
        if selection:
            n0 = self.data["s0_lookup"].select(selection)
            n1 = self.data["s1_lookup"].select(selection)
            values, matrix = calculate_harmonic_tables_from_selections(
                n0, n1, self.coefficients_list
            )
            self._harmonic_values = [values]
            self._matrices = [matrix]
        elif block_selections:
            self._harmonic_values = []
            self._matrices = []
            for sel in block_selections:
                n0 = self.data["s0_lookup"].select(sel)
                n1 = self.data["s1_lookup"].select(sel)
                values, matrix = calculate_harmonic_tables_from_selections(
                    n0, n1, self.coefficients_list
                )
                self._harmonic_values.append(values)
                self._matrices.append(matrix)
        else:
            n0 = self.data["s0_lookup"]
            n1 = self.data["s1_lookup"]
            values, matrix = calculate_harmonic_tables_from_selections(
                n0, n1, self.coefficients_list
            )
            self._harmonic_values = [values]
            self._matrices = [matrix]
        self._n_refl = [val[0].size() for val in self._harmonic_values]

    def _update_reflection_data_speedmode(self, selection=None, block_selections=None):
        if selection:
            sel_sph_harm_table = self.data["sph_harm_table"].select_columns(
                selection.iselection()
            )
            self._harmonic_values = [sel_sph_harm_table.transpose()]
        elif block_selections:
            self._harmonic_values = []
            for sel in block_selections:
                block_sph_harm_table = self.data["sph_harm_table"].select_columns(sel)
                self._harmonic_values.append(block_sph_harm_table.transpose())
        else:
            self._harmonic_values = [self.data["sph_harm_table"].transpose()]
        self._n_refl = [val.n_rows for val in self._harmonic_values]

    def calculate_scales(self, block_id=0):
        """Calculate and return inverse scales for a given block."""
        if self._mode == "speed":
            return self._calculate_scales_and_derivatives_speedmode(
                block_id, derivatives=False
            )
        elif self._mode == "memory":
            return self._calculate_scales_and_derivatives_memorymode(
                block_id, derivatives=False
            )

    def calculate_scales_and_derivatives(self, block_id=0):
        """Calculate and return inverse scales and derivatives for a given block."""
        if self._mode == "speed":
            return self._calculate_scales_and_derivatives_speedmode(block_id)
        elif self._mode == "memory":
            return self._calculate_scales_and_derivatives_memorymode(block_id)

    def _calculate_scales_and_derivatives_speedmode(self, block_id, derivatives=True):
        abs_scale = flex.double(
            self._harmonic_values[block_id].n_rows, 1.0
        )  # Unity term
        for i, col in enumerate(self._harmonic_values[block_id].cols()):
            abs_scale += flex.double(col.as_dense_vector() * self._parameters[i])
        if derivatives:
            return abs_scale, self._harmonic_values[block_id]
        return abs_scale

    def _calculate_scales_and_derivatives_memorymode(self, block_id, derivatives=True):
        abs_scale = flex.double(
            self._harmonic_values[block_id][0].size(), 1.0
        )  # Unity term
        for i, arr in enumerate(
            self._harmonic_values[block_id]
        ):  # iterate over a list of arrays
            abs_scale += arr * self._parameters[i]
        if derivatives:
            return abs_scale, self._matrices[block_id]
        return abs_scale



"""LBFGS refinement engine for cosym analysis."""

from __future__ import annotations

import logging

import scipy.optimize

import scitbx.lbfgs
from scitbx.array_family import flex

logger = logging.getLogger(__name__)


class lbfgs_with_curvs:
    """Minimise a target function using the LBFGS minimiser.

    Implementation of an LBFGS minimiser using curvature information, according
    to the interface defined by :mod:`scitbx.lbfgs`.
    """

    def __init__(self, target, coords, use_curvatures=True, termination_params=None):
        """Initialise an lbfgs_with_curvs object.

        Args:
          target (dials.algorithms.target.Target): The target function to minimise.
          coords (np.ndarray): The starting coordinates for minimisation.
          use_curvatures (bool): Whether or not to use curvature information in the
            minimisation. Defaults to True.
          termination_params (scitbx.lbfgs.termination_parameters):
            Override the default termination parameters for the minimisation.
        """
        self.target = target

        self.x = flex.double(coords)
        self.f = None
        self.g = None

        if use_curvatures:
            self.diag_mode = "always"
        else:
            self.diag_mode = None

        self.minimizer = scitbx.lbfgs.run(
            target_evaluator=self, termination_params=termination_params
        )
        self.coords = self.x.as_numpy_array()

    def compute_functional_gradients_diag(self):
        """Compute the functional, gradients and diagonal.

        Returns:
          tuple: A tuple of the functional, gradients and diagonal, where the
          diagonal is the reciprocal of the curvatures.
        """
        f, g, curvs = self.compute_functional_gradients_and_curvatures()

        # Curvatures of zero will cause a crash, because their inverse is taken.
        assert curvs.all_gt(0.0)
        diags = 1.0 / curvs
        return f, g, diags

    def compute_functional_gradients_and_curvatures(self):
        """Compute the functional, gradients and curvatures.

        Returns:
          tuple: A tuple of the functional, gradients and curvatures.
        """
        x = self.x.as_numpy_array()
        self.f = self.target.compute_functional(x)
        self.g = self.target.compute_gradients(x)
        self.c = self.target.curvatures(x)
        return self.f, flex.double(self.g), flex.double(self.c)

    def compute_functional_and_gradients(self):
        """Compute the functional and gradients.

        Returns:
          tuple: A tuple of the functional and gradients.
        """
        x = self.x.as_numpy_array()
        self.f = self.target.compute_functional(x)
        self.g = self.target.compute_gradients(x)
        return self.f, flex.double(self.g)

    def callback_after_step(self, minimizer):
        """Log progress after each successful step of the minimisation."""
        logger.debug("minimization step: f, iter, nfun:")
        logger.debug(f"{self.f} {minimizer.iter()} {minimizer.nfun()}")


def minimize_scitbx_lbfgs(
    target, coords, use_curvatures=True, max_iterations=100, max_calls=None
):

    termination_params = scitbx.lbfgs.termination_parameters(
        max_iterations=max_iterations,
        max_calls=max_calls,
        traditional_convergence_test=True,
        traditional_convergence_test_eps=1,
        drop_convergence_test_n_test_points=5,
        drop_convergence_test_max_drop_eps=1.0e-5,
        drop_convergence_test_iteration_coefficient=2,
    )
    result = lbfgs_with_curvs(
        target,
        coords,
        use_curvatures=use_curvatures,
        termination_params=termination_params,
    )
    return scipy.optimize.OptimizeResult(
        fun=result.f, jac=result.g, x=result.coords, nfev=result.minimizer.nfun()
    )


def minimize_scipy(
    target, coords, method="L-BFGS-B", max_iterations=None, max_calls=None
):
    """Thin wrapper around scipy.optimize.minimize.

    Args:
      target (dials.algorithms.target.Target): The target function to minimise.
      coords (np.array): The starting coordinates for
        minimisation.
    """
    options = {}
    if max_iterations:
        options.update(maxiter=max_iterations)
    if max_calls:
        options.update(maxfun=max_calls)
    return scipy.optimize.minimize(
        fun=target.compute_functional,
        x0=coords,
        jac=target.compute_gradients,
        method=method,
        options=options,
    )



from __future__ import annotations


class EllipsoidProfileModelExt:
    """An extension class implementing a reciprocal space multivariate normal profile model."""

    name = "ellipsoid"

    default = False

    @staticmethod
    def phil():
        from dials.algorithms.profile_model.ellipsoid.model import phil_scope

        return phil_scope

    @staticmethod
    def algorithm():
        from dials.algorithms.profile_model.ellipsoid.model import EllipsoidProfileModel

        return EllipsoidProfileModel

    @classmethod
    def from_dict(cls, d):
        return cls.algorithm().from_dict(d)



from __future__ import annotations


def generate_phil_scope():
    """
    Generate the phil scope for profile model

    :return: The phil scope
    """
    import dials.extensions

    phil_scope = dials.extensions.ProfileModel.phil_scope()
    return phil_scope


phil_scope = generate_phil_scope()


class ProfileModelFactory:
    """
    Factory for creating profile models
    """

    @staticmethod
    def create(params, experiments, reflections=None):
        """
        Compute or load the profile model.

        :param params: The input phil parameters
        :param experiments: The experiment list
        :param reflections: The reflection table
        :return: The profile model
        """
        import dials.extensions

        Extension = dials.extensions.ProfileModel.load(params.profile.algorithm)
        Algorithm = Extension().algorithm()
        create_profile_model = True
        if hasattr(params, "create_profile_model"):
            create_profile_model = params.create_profile_model
        if (
            create_profile_model
            and reflections is not None
            and "shoebox" in reflections
        ):
            for expr, indices in reflections.iterate_experiments_and_indices(
                experiments
            ):
                expr.profile = Algorithm.create(
                    params.profile,
                    reflections.select(indices),
                    expr.crystal,
                    expr.beam,
                    expr.detector,
                    expr.goniometer,
                    expr.scan,
                )
        else:
            for expr in experiments:
                expr.profile = Algorithm.create(
                    params.profile,
                    None,
                    expr.crystal,
                    expr.beam,
                    expr.detector,
                    expr.goniometer,
                    expr.scan,
                    expr.profile,
                )
        return experiments



"""
Test for dials.assign_experiment_identifiers
"""

from __future__ import annotations

import os

import procrunner

from dxtbx.serialize import load

from dials.array_family import flex


def run_assign_identifiers(pickle_path_list, sequence_path_list, extra_args):
    command = (
        ["dials.assign_experiment_identifiers"]
        + pickle_path_list
        + sequence_path_list
        + extra_args
    )
    print(command)
    procrunner.run(command).check_returncode()
    assert os.path.exists("assigned.expt")
    assert os.path.exists("assigned.refl")


def test_assign_identifiers(dials_data, run_in_tmp_path):
    """Test for dials.assign_experiment_identifiers"""
    pickle_path_list = []
    sequence_path_list = []
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    for i in [20, 25]:
        pickle_path_list.append(data_dir / f"{i}_integrated.pickle")
        sequence_path_list.append(data_dir / f"{i}_integrated_experiments.json")

    run_assign_identifiers(pickle_path_list, sequence_path_list, extra_args=[])

    r = flex.reflection_table.from_file("assigned.refl")
    e = load.experiment_list("assigned.expt", check_format=False)
    r.assert_experiment_identifiers_are_consistent(e)
    assert list(r.experiment_identifiers().values()) != ["", ""]
    assert list(r.experiment_identifiers().keys()) == [0, 1]
    assert list(e.identifiers()) == list(r.experiment_identifiers().values())

    # now run again, with already assigned data
    pickle_path_list = ["assigned.refl"]
    sequence_path_list = ["assigned.expt"]
    run_assign_identifiers(pickle_path_list, sequence_path_list, extra_args=[])

    r = flex.reflection_table.from_file("assigned.refl")
    e = load.experiment_list("assigned.expt", check_format=False)
    r.assert_experiment_identifiers_are_consistent(e)
    assert list(r.experiment_identifiers().values()) != ["", ""]
    assert list(r.experiment_identifiers().keys()) == [0, 1]
    assert list(e.identifiers()) == list(r.experiment_identifiers().values())

    # now run again, with adding more data
    pickle_path_list = ["assigned.refl"]
    sequence_path_list = ["assigned.expt"]
    for i in [30, 35]:
        pickle_path_list.append(data_dir / f"{i}_integrated.pickle")
        sequence_path_list.append(data_dir / f"{i}_integrated_experiments.json")

    run_assign_identifiers(
        pickle_path_list, sequence_path_list, extra_args=["identifiers=0 5 10 15"]
    )

    r = flex.reflection_table.from_file("assigned.refl")
    e = load.experiment_list("assigned.expt", check_format=False)
    r.assert_experiment_identifiers_are_consistent(e)
    assert list(r.experiment_identifiers().values()) == ["0", "5", "10", "15"]
    assert list(r.experiment_identifiers().keys()) == [0, 1, 2, 3]
    assert list(e.identifiers()) == ["0", "5", "10", "15"]



from __future__ import annotations

import iotbx.detectors

# store default ImageFactory function
defaultImageFactory = iotbx.detectors.ImageFactory


def SlipViewerImageFactory(filename):
    try:
        return NpyImageFactory(filename)
    except Exception:
        return defaultImageFactory(filename)


# Use the dxtbx class as it handles all possible variance of NPY images
def NpyImageFactory(filename):
    from dxtbx.format.FormatPYunspecified import FormatPYunspecified

    img = FormatPYunspecified(filename)
    return img.get_detectorbase()



from __future__ import annotations

import cmath
import concurrent.futures
import copy
import itertools
import logging
import math
import random

import iotbx.phil
import libtbx.introspection
from libtbx.test_utils import approx_equal
from libtbx.utils import plural_s
from rstbx.dps_core import Direction, Directional_FFT
from rstbx.indexing_api import dps_extended
from rstbx.indexing_api.lattice import DPS_primitive_lattice
from rstbx.phil.phil_preferences import indexing_api_defs
from scitbx import matrix
from scitbx.array_family import flex
from scitbx.simplex import simplex_opt

import dials.util
from dials.algorithms.indexing.indexer import find_max_cell
from dials.util import Sorry, log
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.slice import slice_reflections

logger = logging.getLogger("dials.command_line.search_beam_position")

help_message = """

Search for a better beam centre using the results of spot finding. Based on
method of Sauter et al., J. Appl. Cryst. 37, 399-409 (2004).

Examples::

  dials.search_beam_position imported.expt strong.refl
"""

phil_scope = iotbx.phil.parse(
    """
nproc = Auto
  .type = int(value_min=1)
plot_search_scope = False
  .type = bool
max_cell = None
  .type = float
  .help = "Known max cell (otherwise will compute from spot positions)"
image_range = None
  .help = "The range of images to use in indexing. Number of arguments"
    "must be a factor of two. Specifying \"0 0\" will use all images"
    "by default. The given range follows C conventions"
    "(e.g. j0 <= j < j1)."
  .type = ints(size=2)
  .multiple = True
max_reflections = 10000
  .type = int(value_min=1)
  .help = "Maximum number of reflections to use in the search for better"
          "experimental model. If the number of input reflections is greater"
          "then a random subset of reflections will be used."
mm_search_scope = 4.0
  .help = "Global radius of origin offset search."
  .type = float(value_min=0)
wide_search_binning = 2
  .help = "Modify the coarseness of the wide grid search for the beam centre."
  .type = float(value_min=0)
n_macro_cycles = 1
  .type = int
  .help = "Number of macro cycles for an iterative beam centre search."
d_min = None
  .type = float(value_min=0)

seed = 42
  .type = int(value_min=0)

output {
  experiments = optimised.expt
    .type = path
  log = "dials.search_beam_position.log"
    .type = str
}
"""
)


def optimize_origin_offset_local_scope(
    experiments,
    reflection_lists,
    solution_lists,
    amax_lists,
    mm_search_scope=4,
    wide_search_binning=1,
    plot_search_scope=False,
):
    """Local scope: find the optimal origin-offset closest to the current overall detector position
    (local minimum, simple minimization)"""

    beam = experiments[0].beam
    s0 = matrix.col(beam.get_s0())
    # construct two vectors that are perpendicular to the beam.  Gives a basis for refining beam
    axis = matrix.col((1, 0, 0))
    beamr0 = s0.cross(axis).normalize()
    beamr1 = beamr0.cross(s0).normalize()
    beamr2 = beamr1.cross(s0).normalize()

    assert approx_equal(s0.dot(beamr1), 0.0)
    assert approx_equal(s0.dot(beamr2), 0.0)
    assert approx_equal(beamr2.dot(beamr1), 0.0)
    # so the orthonormal vectors are s0, beamr1 and beamr2

    if mm_search_scope:
        plot_px_sz = experiments[0].detector[0].get_pixel_size()[0]
        plot_px_sz *= wide_search_binning
        grid = max(1, int(mm_search_scope / plot_px_sz))
        widegrid = 2 * grid + 1

        def get_experiment_score_for_coord(x, y):
            new_origin_offset = x * plot_px_sz * beamr1 + y * plot_px_sz * beamr2
            return sum(
                _get_origin_offset_score(
                    new_origin_offset,
                    solution_lists[i],
                    amax_lists[i],
                    reflection_lists[i],
                    experiment,
                )
                for i, experiment in enumerate(experiments)
            )

        scores = flex.double(
            get_experiment_score_for_coord(x, y)
            for y in range(-grid, grid + 1)
            for x in range(-grid, grid + 1)
        )

        def igrid(x):
            return x - (widegrid // 2)

        idxs = [igrid(i) * plot_px_sz for i in range(widegrid)]

        # if there are several similarly high scores, then choose the closest
        # one to the current beam centre
        potential_offsets = flex.vec3_double()
        if scores.all_eq(0):
            raise Sorry("No valid scores")
        sel = scores > (0.9 * flex.max(scores))
        for i in sel.iselection():
            offset = (idxs[i % widegrid]) * beamr1 + (idxs[i // widegrid]) * beamr2
            potential_offsets.append(offset.elems)
            # print offset.length(), scores[i]
        wide_search_offset = matrix.col(
            potential_offsets[flex.min_index(potential_offsets.norms())]
        )

    else:
        wide_search_offset = None

    # Do a simplex minimization
    class simplex_minimizer:
        def __init__(self, wide_search_offset):
            self.n = 2
            self.wide_search_offset = wide_search_offset
            self.optimizer = simplex_opt(
                dimension=self.n,
                matrix=[flex.random_double(self.n) for _ in range(self.n + 1)],
                evaluator=self,
                tolerance=1e-7,
            )
            self.x = self.optimizer.get_solution()
            self.offset = self.x[0] * 0.2 * beamr1 + self.x[1] * 0.2 * beamr2
            if self.wide_search_offset is not None:
                self.offset += self.wide_search_offset

        def target(self, vector):
            trial_origin_offset = vector[0] * 0.2 * beamr1 + vector[1] * 0.2 * beamr2
            if self.wide_search_offset is not None:
                trial_origin_offset += self.wide_search_offset
            target = 0
            for i, experiment in enumerate(experiments):
                target -= _get_origin_offset_score(
                    trial_origin_offset,
                    solution_lists[i],
                    amax_lists[i],
                    reflection_lists[i],
                    experiment,
                )
            return target

    new_offset = simplex_minimizer(wide_search_offset).offset

    if plot_search_scope:
        plot_px_sz = experiments[0].get_detector()[0].get_pixel_size()[0]
        grid = max(1, int(mm_search_scope / plot_px_sz))
        scores = flex.double()
        for y in range(-grid, grid + 1):
            for x in range(-grid, grid + 1):
                new_origin_offset = x * plot_px_sz * beamr1 + y * plot_px_sz * beamr2
                score = 0
                for i, experiment in enumerate(experiments):
                    score += _get_origin_offset_score(
                        new_origin_offset,
                        solution_lists[i],
                        amax_lists[i],
                        reflection_lists[i],
                        experiment,
                    )
                scores.append(score)

        def show_plot(widegrid, excursi):
            excursi.reshape(flex.grid(widegrid, widegrid))
            idx_max = flex.max_index(excursi)

            def igrid(x):
                return x - (widegrid // 2)

            idxs = [igrid(i) * plot_px_sz for i in range(widegrid)]

            from matplotlib import pyplot as plt

            plt.figure()
            CS = plt.contour(
                [igrid(i) * plot_px_sz for i in range(widegrid)],
                [igrid(i) * plot_px_sz for i in range(widegrid)],
                excursi.as_numpy_array(),
            )
            plt.clabel(CS, inline=1, fontsize=10, fmt="%6.3f")
            plt.title("Wide scope search for detector origin offset")
            plt.scatter([0.0], [0.0], color="g", marker="o")
            plt.scatter([new_offset[0]], [new_offset[1]], color="r", marker="*")
            plt.scatter(
                [idxs[idx_max % widegrid]],
                [idxs[idx_max // widegrid]],
                color="k",
                marker="s",
            )
            plt.axes().set_aspect("equal")
            plt.xlabel("offset (mm) along beamr1 vector")
            plt.ylabel("offset (mm) along beamr2 vector")
            plt.savefig("search_scope.png")

            # changing value
            trial_origin_offset = (idxs[idx_max % widegrid]) * beamr1 + (
                idxs[idx_max // widegrid]
            ) * beamr2
            return trial_origin_offset

        show_plot(widegrid=2 * grid + 1, excursi=scores)

    new_experiments = copy.deepcopy(experiments)
    for expt in new_experiments:
        expt.detector = dps_extended.get_new_detector(expt.detector, new_offset)
    return new_experiments


def _get_origin_offset_score(
    trial_origin_offset, solutions, amax, spots_mm, experiment
):
    trial_detector = dps_extended.get_new_detector(
        experiment.detector, trial_origin_offset
    )
    experiment = copy.copy(experiment)
    experiment.detector = trial_detector

    # Key point for this is that the spots must correspond to detector
    # positions not to the correct RS position => reset any fixed rotation
    # to identity

    experiment.goniometer.set_fixed_rotation((1, 0, 0, 0, 1, 0, 0, 0, 1))
    spots_mm.map_centroids_to_reciprocal_space([experiment])
    return _sum_score_detail(spots_mm["rlp"], solutions, amax=amax)


def _sum_score_detail(reciprocal_space_vectors, solutions, granularity=None, amax=None):
    """Evaluates the probability that the trial value of (S0_vector | origin_offset) is correct,
    given the current estimate and the observations.  The trial value comes through the
    reciprocal space vectors, and the current estimate comes through the short list of
    DPS solutions. Actual return value is a sum of NH terms, one for each DPS solution, each ranging
    from -1.0 to 1.0"""

    nh = min(solutions.size(), 20)  # extended API
    sum_score = 0.0
    for t in range(nh):
        dfft = Directional_FFT(
            angle=Direction(solutions[t]),
            xyzdata=reciprocal_space_vectors,
            granularity=5.0,
            amax=amax,  # extended API XXX These values have to come from somewhere!
            F0_cutoff=11,
        )
        kval = dfft.kval()
        kmax = dfft.kmax()
        kval_cutoff = reciprocal_space_vectors.size() / 4.0
        if kval > kval_cutoff:
            ff = dfft.fft_result
            kbeam = ((-dfft.pmin) / dfft.delta_p) + 0.5
            Tkmax = cmath.phase(ff[kmax])
            backmax = math.cos(
                Tkmax + (2 * math.pi * kmax * kbeam / (2 * ff.size() - 1))
            )
            ### Here it should be possible to calculate a gradient.
            ### Then minimize with respect to two coordinates.  Use lbfgs?  Have second derivatives?
            ### can I do something local to model the cosine wave?
            ### direction of wave travel.  Period. phase.
            sum_score += backmax
    return sum_score


def run_dps(experiment, spots_mm, max_cell):
    # max_cell: max possible cell in Angstroms; set to None, determine from data
    # recommended_grid_sampling_rad: grid sampling in radians; guess for now

    horizon_phil = iotbx.phil.parse(input_string=indexing_api_defs).extract()
    DPS = DPS_primitive_lattice(
        max_cell=max_cell, recommended_grid_sampling_rad=None, horizon_phil=horizon_phil
    )

    DPS.S0_vector = matrix.col(experiment.beam.get_s0())
    DPS.inv_wave = 1.0 / experiment.beam.get_wavelength()
    if experiment.goniometer is None:
        DPS.axis = matrix.col((1, 0, 0))
    else:
        DPS.axis = matrix.col(experiment.goniometer.get_rotation_axis())
    DPS.set_detector(experiment.detector)

    # transform input into what DPS needs
    # i.e., construct a flex.vec3 double consisting of mm spots, phi in degrees
    data = flex.vec3_double()
    for spot in spots_mm.rows():
        data.append(
            (
                spot["xyzobs.mm.value"][0],
                spot["xyzobs.mm.value"][1],
                spot["xyzobs.mm.value"][2] * 180.0 / math.pi,
            )
        )

    logger.info("Running DPS using %i reflections", len(data))

    DPS.index(
        raw_spot_input=data,
        panel_addresses=flex.int(s["panel"] for s in spots_mm.rows()),
    )
    solutions = DPS.getSolutions()

    logger.info(
        "Found %i solution%s with max unit cell %.2f Angstroms.",
        len(solutions),
        plural_s(len(solutions))[1],
        DPS.amax,
    )

    # There must be at least 3 solutions to make a set, otherwise return empty result
    if len(solutions) < 3:
        return {}
    return {"solutions": flex.vec3_double(s.dvec for s in solutions), "amax": DPS.amax}


def discover_better_experimental_model(
    experiments,
    reflections,
    params,
    nproc=1,
    d_min=None,
    mm_search_scope=4.0,
    wide_search_binning=1,
    plot_search_scope=False,
):
    assert len(experiments) == len(reflections)
    assert len(experiments) > 0

    refl_lists = []
    max_cell_list = []

    # The detector/beam of the first experiment is used to define the basis for the
    # optimisation, so assert that the beam intersects with the detector
    detector = experiments[0].detector
    beam = experiments[0].beam
    beam_panel = detector.get_panel_intersection(beam.get_s0())
    if beam_panel == -1:
        raise Sorry("input beam does not intersect detector")

    for expt, refl in zip(experiments, reflections):
        refl = copy.deepcopy(refl)
        refl["imageset_id"] = flex.int(len(refl), 0)
        refl.centroid_px_to_mm([expt])
        refl.map_centroids_to_reciprocal_space([expt])

        if d_min is not None:
            d_spacings = 1 / refl["rlp"].norms()
            sel = d_spacings > d_min
            refl = refl.select(sel)

        # derive a max_cell from mm spots
        if params.max_cell is None:
            max_cell = find_max_cell(
                refl, max_cell_multiplier=1.3, step_size=45
            ).max_cell
            max_cell_list.append(max_cell)

        if params.max_reflections is not None and refl.size() > params.max_reflections:
            logger.info(
                "Selecting subset of %i reflections for analysis",
                params.max_reflections,
            )
            perm = flex.random_permutation(refl.size())
            sel = perm[: params.max_reflections]
            refl = refl.select(sel)

        refl_lists.append(refl)

    if params.max_cell is None:
        max_cell = flex.median(flex.double(max_cell_list))
    else:
        max_cell = params.max_cell

    with concurrent.futures.ProcessPoolExecutor(max_workers=nproc) as pool:
        solution_lists = []
        amax_list = []
        for result in pool.map(
            run_dps, experiments, refl_lists, itertools.repeat(max_cell)
        ):
            if result.get("solutions"):
                solution_lists.append(result["solutions"])
                amax_list.append(result["amax"])

    if not solution_lists:
        raise Sorry("No solutions found")

    new_experiments = optimize_origin_offset_local_scope(
        experiments,
        refl_lists,
        solution_lists,
        amax_list,
        mm_search_scope=mm_search_scope,
        wide_search_binning=wide_search_binning,
        plot_search_scope=plot_search_scope,
    )
    new_detector = new_experiments[0].detector
    old_panel, old_beam_centre = detector.get_ray_intersection(beam.get_s0())
    new_panel, new_beam_centre = new_detector.get_ray_intersection(beam.get_s0())

    old_beam_centre_px = detector[old_panel].millimeter_to_pixel(old_beam_centre)
    new_beam_centre_px = new_detector[new_panel].millimeter_to_pixel(new_beam_centre)

    logger.info(
        "Old beam centre: %.2f, %.2f mm" % old_beam_centre
        + " (%.1f, %.1f px)" % old_beam_centre_px
    )
    logger.info(
        "New beam centre: %.2f, %.2f mm" % new_beam_centre
        + " (%.1f, %.1f px)" % new_beam_centre_px
    )
    logger.info(
        "Shift: %.2f, %.2f mm"
        % (matrix.col(old_beam_centre) - matrix.col(new_beam_centre)).elems
        + " (%.1f, %.1f px)"
        % (matrix.col(old_beam_centre_px) - matrix.col(new_beam_centre_px)).elems
    )
    return new_experiments


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.search_beam_position [options] imported.expt strong.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=False)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 or len(reflections) == 0:
        parser.print_help()
        exit(0)

    # Configure the logging
    log.config(logfile=params.output.log)

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    if params.seed is not None:
        flex.set_random_seed(params.seed)
        random.seed(params.seed)

    if params.nproc is libtbx.Auto:
        params.nproc = libtbx.introspection.number_of_processors()

    imagesets = experiments.imagesets()
    # Split all the refln tables by ID, corresponding to the respective imagesets
    reflections = [
        refl_unique_id
        for refl in reflections
        for refl_unique_id in refl.split_by_experiment_id()
    ]

    assert len(imagesets) > 0
    assert len(reflections) == len(imagesets)

    if params.image_range is not None and len(params.image_range) > 0:
        reflections = [
            slice_reflections(refl, params.image_range) for refl in reflections
        ]

    for i in range(params.n_macro_cycles):
        if params.n_macro_cycles > 1:
            logger.info("Starting macro cycle %i", i + 1)
        experiments = discover_better_experimental_model(
            experiments,
            reflections,
            params,
            nproc=params.nproc,
            d_min=params.d_min,
            mm_search_scope=params.mm_search_scope,
            wide_search_binning=params.wide_search_binning,
            plot_search_scope=params.plot_search_scope,
        )
        logger.info("")

    logger.info("Saving optimised experiments to %s", params.output.experiments)
    experiments.as_file(params.output.experiments)


if __name__ == "__main__":
    run()
